#!/usr/bin/env python3
import os
import tempfile
import shutil
import subprocess
import argparse
import re


def parse_version(version_string):
    match = re.fullmatch(r'v(\d+)\.(\d+)\.(\d+)', version_string)
    if not match:
        raise ValueError("Invalid version string")
    return match.group(1), match.group(2), match.group(3)


argument_parser = argparse.ArgumentParser(
    description="Helper to import libcbor as external dependency.")

argument_parser.add_argument("--version",
                             required=True, help="Version string to import")

args = argument_parser.parse_args()
major_version, minor_version, patch_version = parse_version(args.version)

GENERATED_NOTES = """/**
 * DO NOT DIRECTLY MODIFY THIS FILE:
 *
 * The code in this file is generated from scripts/import_libcbor.py
 * and any modifications should be in there.
 */
"""

CBOR_EXPORT_H = """
#ifndef CBOR_EXPORT_H
#define CBOR_EXPORT_H

/* Don't export anything from libcbor */
#define CBOR_EXPORT

#endif /* CBOR_EXPORT_H */
"""

CONFIGURATION_H = f"""
#ifndef LIBCBOR_CONFIGURATION_H
#define LIBCBOR_CONFIGURATION_H

#define CBOR_MAJOR_VERSION {major_version}
#define CBOR_MINOR_VERSION {minor_version}
#define CBOR_PATCH_VERSION {patch_version}

#define CBOR_BUFFER_GROWTH 2
#define CBOR_MAX_STACK_SIZE 2048
#define CBOR_PRETTY_PRINTER 1

#if defined(_MSC_VER)
#    define CBOR_RESTRICT_SPECIFIER
#else
#    define CBOR_RESTRICT_SPECIFIER restrict
#endif

#define CBOR_INLINE_SPECIFIER

/* Ignore the compiler warnings for libcbor. */
#ifdef _MSC_VER
#    pragma warning(disable : 4028)
#    pragma warning(disable : 4189)
#    pragma warning(disable : 4715)
#    pragma warning(disable : 4232)
#    pragma warning(disable : 4068)
#    pragma warning(disable : 4244)
#    pragma warning(disable : 4701)
#    pragma warning(disable : 4703)
#endif

#ifdef __clang__
#    pragma clang diagnostic ignored "-Wreturn-type"
#elif defined(__GNUC__)
#    pragma GCC diagnostic ignored "-Wreturn-type"
#    pragma GCC diagnostic ignored "-Wunknown-pragmas"
#    pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
#endif

#endif // LIBCBOR_CONFIGURATION_H
"""


# Create a temporary directory for cloning the repository
temp_repo_dir = tempfile.mkdtemp()

try:
    # Clone the repository into the temporary directory
    repo_url = "https://github.com/PJK/libcbor.git"
    clone_command = f"git clone {repo_url} {temp_repo_dir}"
    subprocess.run(clone_command, shell=True, check=True)
    subprocess.run(["git", "checkout", "tags/" + args.version],
                   cwd=temp_repo_dir, check=True)
    version_number = args.version[1:]  # Remove 'v' prefix

    # Create a separate folder for the copied files
    base_dir = os.path.join(os.path.dirname(
        __file__), "..")
    output_dir = os.path.join(
        base_dir, "source", "external", "libcbor")
    shutil.rmtree(output_dir, ignore_errors=True)
    os.makedirs(output_dir, exist_ok=True)

    # Copy files ending with .c and .h from the /src directory
    src_dir = os.path.join(temp_repo_dir, "src")
    for root, dirs, files in os.walk(src_dir):
        dir_name = os.path.basename(root)
        for file in files:
            if file.endswith((".c", ".h")):
                # copy the source code to ../source/external/libcbor
                src_file = os.path.join(root, file)
                rel_path = os.path.relpath(src_file, src_dir)
                dest_file = os.path.join(output_dir, rel_path)
                os.makedirs(os.path.dirname(dest_file), exist_ok=True)
                shutil.copy(src_file, dest_file)

    # Use our customized configurations
    with open(os.path.join(output_dir, "cbor/cbor_export.h"), "w") as file:
        file.write(GENERATED_NOTES)
        file.write(CBOR_EXPORT_H)
    with open(os.path.join(output_dir, "cbor/configuration.h"), "w") as file:
        file.write(GENERATED_NOTES)
        file.write(CONFIGURATION_H)

    # Update the version number in THIRD-PARTY-LICENSES.txt
    license_path = os.path.join(base_dir, "THIRD-PARTY-LICENSES.txt")
    with open(license_path, 'r') as f:
        content = f.read()

    # Update the version number using regex for more accurate replacement
    pattern = r'\*\* libcbor; version \d+\.\d+\.\d+ --'
    replacement = f'** libcbor; version {version_number} --'
    content = re.sub(pattern, replacement, content)

    with open(license_path, 'w') as f:
        f.write(content)

    print(f"Updated libcbor to version {version_number}")

except Exception as e:
    print(f"An error occurred: {e}")

finally:
    # Remove the temporary directory
    shutil.rmtree(temp_repo_dir, ignore_errors=True)



# Contains all of the metric reporting functions for the Canary Wrappers

# Needs to be installed prior to running
import psutil


cache_cpu_psutil_process = None
def get_metric_total_cpu_usage(psutil_process : psutil.Process):
    global cache_cpu_psutil_process

    try:
        if (psutil_process == None):
            print ("ERROR - No psutil.process passed! Cannot gather metric!", flush=True)
            return None
        # We always need to skip the first CPU poll
        if (cache_cpu_psutil_process != psutil_process):
            psutil.cpu_percent(interval=None)
            cache_cpu_psutil_process = psutil_process
            return None
        return psutil.cpu_percent(interval=None)
    except Exception as e:
        print ("ERROR - exception occurred gathering metrics!")
        print ("Exception: " + str(e), flush=True)
        return None

# Note: This value is in BYTES.
def get_metric_total_memory_usage_value(psutil_process : psutil.Process):
    try:
        if (psutil_process == None):
            print ("ERROR - No psutil.process passed! Cannot gather metric!", flush=True)
            return None
        return psutil.virtual_memory()[3]
    except Exception as e:
        print ("ERROR - exception occurred gathering metrics!")
        print ("Exception: " + str(e), flush=True)
        return None


def get_metric_total_memory_usage_percent(psutil_process : psutil.Process):
    try:
        if (psutil_process == None):
            print ("ERROR - No psutil.process passed! Cannot gather metric!", flush=True)
            return None
        return psutil.virtual_memory()[2]
    except Exception as e:
        print ("ERROR - exception occurred gathering metrics!")
        print ("Exception: " + str(e), flush=True)
        return None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from common import Protocols
from providers import S2N
from global_flags import get_flag, S2N_FIPS_MODE


def to_bytes(val):
    return bytes(str(val).encode("utf-8"))


def to_string(val: bytes):
    return val.decode(encoding="ascii", errors="backslashreplace")


def get_expected_s2n_version(protocol, provider):
    """
    s2nd and s2nc print a number for the negotiated TLS version.

    provider is s2n's peer. If s2n tries to speak to s2n < tls13,
    tls12 is always chosen. This is true even when the requested
    protocol is less than tls12.
    """
    if provider == S2N and protocol != Protocols.TLS13:
        version = "33"
    else:
        version = protocol.value

    return version


def get_expected_openssl_version(protocol):
    return {
        Protocols.TLS10.value: "TLSv1",
        Protocols.TLS11.value: "TLSv1.1",
        Protocols.TLS12.value: "TLSv1.2",
        Protocols.TLS13.value: "TLSv1.3",
    }.get(protocol.value)


def get_expected_gnutls_version(protocol):
    return {
        Protocols.TLS10.value: "TLS1.0",
        Protocols.TLS11.value: "TLS1.1",
        Protocols.TLS12.value: "TLS1.2",
        Protocols.TLS13.value: "TLS1.3",
    }.get(protocol.value)


def get_parameter_name(item):
    if isinstance(item, type):
        return item.__name__
    return str(item)


def invalid_test_parameters(*args, **kwargs):
    """
    Determine if the parameters chosen for a test makes sense.
    This function returns True or False, indicating whether a
    test should be "deselected" based on the arguments.
    """
    protocol = kwargs.get("protocol")
    provider = kwargs.get("provider")
    other_provider = kwargs.get("other_provider")
    certificate = kwargs.get("certificate")
    client_certificate = kwargs.get("client_certificate")
    cipher = kwargs.get("cipher")
    curve = kwargs.get("curve")
    signature = kwargs.get("signature")

    providers = [provider_ for provider_ in [provider, other_provider] if provider_]
    # Always consider S2N
    providers.append(S2N)

    certificates = [cert for cert in [certificate, client_certificate] if cert]

    # Older versions do not support RSA-PSS-PSS certificates
    if protocol and protocol < Protocols.TLS12:
        if client_certificate and client_certificate.algorithm == "RSAPSS":
            return True
        if certificate and certificate.algorithm == "RSAPSS":
            return True

    for provider_ in providers:
        if not provider_.supports_protocol(protocol):
            return True

        for certificate_ in certificates:
            if not provider_.supports_certificate(certificate_):
                return True

    if cipher is not None:
        # If the selected protocol doesn't allow the cipher, don't test
        if protocol is not None:
            if cipher.min_version > protocol:
                return True

            # Ciphersuites prior to TLS13 can not be used with TLS13
            # https://wiki.openssl.org/index.php/TLS1.3#Differences_with_TLS1.2_and_below
            if protocol is Protocols.TLS13 and cipher.min_version < protocol:
                return True

        for provider_ in providers:
            if not provider_.supports_cipher(cipher, with_curve=curve):
                return True

        if get_flag(S2N_FIPS_MODE):
            if not cipher.fips:
                return True

    # If we are using a cipher that depends on a specific certificate algorithm
    # deselect the test if the wrong certificate is used.
    if certificate is not None:
        if cipher is not None and certificate.compatible_with_cipher(cipher) is False:
            return True

    # If the curve is specified, then all signatures must use that curve
    if curve:
        if certificate and not certificate.compatible_with_curve(curve):
            return True
        if client_certificate and not client_certificate.compatible_with_curve(curve):
            return True

    # Prevent situations like using X25519 with TLS1.2
    if curve is not None:
        if protocol is not None and curve.min_protocol > protocol:
            return True

    if signature is not None:
        for provider_ in providers:
            if provider_.supports_signature(signature) is False:
                return True

    return False



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
import os
import pytest

from configuration import (
    available_ports,
    ALL_TEST_CURVES,
    MINIMAL_TEST_CERTS,
    TLS13_CIPHERS,
)
from common import ProviderOptions, Protocols, Curves, data_bytes
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N as S2NBase, OpenSSL as OpenSSLBase
from utils import invalid_test_parameters, get_parameter_name, to_bytes

from test_hello_retry_requests import S2N_HRR_MARKER

TICKET_FILE = "ticket"
EARLY_DATA_FILE = "early_data"

MAX_EARLY_DATA = 500  # Arbitrary largish number
DATA_TO_SEND = data_bytes(500)  # Arbitrary large number

NUM_RESUMES = 5  # Hardcoded for s2nc --reconnect
NUM_CONNECTIONS = NUM_RESUMES + 1  # resumes + initial

S2N_DEFAULT_CURVE = Curves.X25519
# We have no plans to support this curve any time soon
S2N_UNSUPPORTED_CURVE = "X448"
S2N_HRR_CURVES = list(curve for curve in ALL_TEST_CURVES if curve != S2N_DEFAULT_CURVE)

S2N_EARLY_DATA_MARKER = to_bytes("WITH_EARLY_DATA")
S2N_EARLY_DATA_RECV_MARKER = "Early Data received: "
S2N_EARLY_DATA_STATUS_MARKER = "Early Data status: {status}"
S2N_EARLY_DATA_ACCEPTED_MARKER = S2N_EARLY_DATA_STATUS_MARKER.format(status="ACCEPTED")
S2N_EARLY_DATA_REJECTED_MARKER = S2N_EARLY_DATA_STATUS_MARKER.format(status="REJECTED")
S2N_EARLY_DATA_NOT_REQUESTED_MARKER = S2N_EARLY_DATA_STATUS_MARKER.format(
    status="NOT REQUESTED"
)


class S2N(S2NBase):
    def __init__(self, options: ProviderOptions):
        S2NBase.__init__(self, options)

    def setup_client(self):
        cmd_line = S2NBase.setup_client(self)
        early_data_file = self.options.early_data_file
        if early_data_file and os.path.exists(early_data_file):
            cmd_line.extend(["--early-data", early_data_file])
        return cmd_line

    def setup_server(self):
        cmd_line = S2NBase.setup_server(self)
        cmd_line.extend(["--max-early-data", self.options.max_early_data])
        return cmd_line


class OpenSSL(OpenSSLBase):
    def __init__(self, options: ProviderOptions):
        OpenSSLBase.__init__(self, options)

    def setup_client(self):
        cmd_line = OpenSSLBase.setup_client(self)
        early_data_file = self.options.early_data_file
        if early_data_file and os.path.exists(early_data_file):
            cmd_line.extend(["-early_data", early_data_file])
        ticket_file = self.options.ticket_file
        if ticket_file:
            if os.path.exists(ticket_file):
                cmd_line.extend(["-sess_in", ticket_file])
            else:
                cmd_line.extend(["-sess_out", self.options.ticket_file])
        return cmd_line

    def setup_server(self):
        cmd_line = OpenSSLBase.setup_server(self)
        if self.options.max_early_data > 0:
            cmd_line.extend(["-early_data"])
        return cmd_line


# The `-reconnect` option is broken for TLS1.3 in OpenSSL s_client: https://github.com/openssl/openssl/issues/8517
# The `-sess_in`/`-sess_out` options can be used instead, but don't have an s2nc equivalent.
# As we add more providers, we may need both a `-reconnect`-like and a `-sess_in/out`-like S2N server test,
# but for now we can just use `-sess_in/out` and cover the S2N->S2N case in the S2N client tests.
CLIENT_PROVIDERS = [OpenSSL]
SERVER_PROVIDERS = [OpenSSL, S2N]


def get_early_data_bytes(file_path, early_data_size):
    early_data = data_bytes(early_data_size)
    with open(file_path, "wb") as fout:
        fout.write(early_data)
    return early_data


def get_ticket_from_s2n_server(options, managed_process, provider, certificate):  # noqa: F811
    port = next(available_ports)

    """
    Generally clients start checking for stdin EoF to exit as soon as they finish the handshake.
    To make sure the client reliably receives the post-handshake NST,
    do NOT indicate stdin EoF until after some data has been received from the server.
    """
    close_marker_bytes = data_bytes(10)

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode
    client_options.port = port

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode
    server_options.port = port
    server_options.key = certificate.key
    server_options.cert = certificate.cert
    server_options.data_to_send = close_marker_bytes

    assert not os.path.exists(options.ticket_file)

    s2n_server = managed_process(
        S2N, server_options, send_marker=S2N.get_send_marker(), timeout=10
    )
    client = managed_process(
        provider, client_options, close_marker=str(close_marker_bytes), timeout=10
    )

    for results in s2n_server.get_results():
        results.assert_success()

    for results in client.get_results():
        results.assert_success()

    assert os.path.exists(options.ticket_file)


def test_nothing():
    """
    Sometimes the early data test parameters in combination with the s2n libcrypto
    results in no test cases existing. In this case, pass a nothing test to avoid
    marking the entire codebuild run as failed.
    """
    assert True


"""
Basic S2N server happy case.

We make one full connection to get a session ticket with early data enabled,
then another resumption connection with early data.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", CLIENT_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize(
    "early_data_size",
    [int(MAX_EARLY_DATA / 2), int(MAX_EARLY_DATA - 1), MAX_EARLY_DATA, 1],
)
def test_s2n_server_with_early_data(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    curve,
    certificate,
    protocol,
    provider,
    other_provider,
    early_data_size,
):
    ticket_file = str(tmp_path / TICKET_FILE)
    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    early_data = get_early_data_bytes(early_data_file, early_data_size)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        data_to_send=DATA_TO_SEND,
    )
    options.ticket_file = ticket_file
    options.early_data_file = early_data_file
    options.max_early_data = MAX_EARLY_DATA

    get_ticket_from_s2n_server(options, managed_process, provider, certificate)

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode

    s2n_server = managed_process(S2N, server_options, timeout=10)
    client = managed_process(provider, client_options, timeout=10)

    for results in client.get_results():
        results.assert_success()

    for results in s2n_server.get_results():
        results.assert_success()
        assert S2N_EARLY_DATA_MARKER in results.stdout
        assert (to_bytes(S2N_EARLY_DATA_RECV_MARKER) + early_data) in results.stdout
        assert to_bytes(S2N_EARLY_DATA_ACCEPTED_MARKER) in results.stdout
        assert DATA_TO_SEND in results.stdout


"""
Basic S2N client happy case.

The S2N client tests session resumption by repeatedly reconnecting.
That means we don't need to manually perform the initial full connection, and there is no external ticket file.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", SERVER_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize(
    "early_data_size",
    [int(MAX_EARLY_DATA / 2), int(MAX_EARLY_DATA - 1), MAX_EARLY_DATA, 1],
)
def test_s2n_client_with_early_data(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    certificate,
    protocol,
    provider,
    other_provider,
    early_data_size,
):
    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    early_data = get_early_data_bytes(early_data_file, early_data_size)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        reconnect=True,
    )
    options.ticket_file = None
    options.early_data_file = early_data_file
    options.max_early_data = MAX_EARLY_DATA

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode
    server_options.key = certificate.key  # Required for the initial connection
    server_options.cert = certificate.cert  # Required for the initial connection
    server_options.reconnects_before_exit = NUM_CONNECTIONS

    server = managed_process(provider, server_options, timeout=10)
    s2n_client = managed_process(S2N, client_options, timeout=10)

    for results in s2n_client.get_results():
        results.assert_success()
        assert S2N_EARLY_DATA_MARKER in results.stdout
        assert (
            results.stdout.count(to_bytes(S2N_EARLY_DATA_ACCEPTED_MARKER))
            == NUM_RESUMES
        )

    for results in server.get_results():
        results.assert_success()
        assert results.stdout.count(early_data) == NUM_RESUMES


"""
Verify that the S2N client doesn't request early data when a server doesn't support early data.

We repeatedly reconnect with max_early_data set to 0. This is basically a test from
test_session_resumption but with validation that no early data is sent.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", SERVER_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
def test_s2n_client_without_early_data(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    certificate,
    protocol,
    provider,
    other_provider,
):
    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    early_data = get_early_data_bytes(early_data_file, MAX_EARLY_DATA)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        reconnect=True,
    )
    options.ticket_file = None
    options.early_data_file = early_data_file
    options.max_early_data = 0

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode
    server_options.key = certificate.key  # Required for the initial connection
    server_options.cert = certificate.cert  # Required for the initial connection
    server_options.reconnects_before_exit = NUM_CONNECTIONS

    server = managed_process(provider, server_options, timeout=10)
    s2n_client = managed_process(S2N, client_options, timeout=10)

    for results in server.get_results():
        results.assert_success()
        assert early_data not in results.stdout

    for results in s2n_client.get_results():
        results.assert_success()
        assert S2N_EARLY_DATA_MARKER not in results.stdout
        assert (
            results.stdout.count(to_bytes(S2N_EARLY_DATA_NOT_REQUESTED_MARKER))
            == NUM_CONNECTIONS
        )


"""
Test the S2N server rejecting early data.

We do this by disabling early data on the server after the ticket is issued.
When the client attempts to use the ticket to send early data, the server rejects the attempt.

We can't perform an S2N client version of this test because the S2N client performs its hardcoded
reconnects automatically, without any mechanism to modify the connection in between.
"""


@pytest.mark.flaky(reruns=5)
@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", CLIENT_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize(
    "early_data_size",
    [int(MAX_EARLY_DATA / 2), int(MAX_EARLY_DATA - 1), MAX_EARLY_DATA, 1],
)
def test_s2n_server_with_early_data_rejected(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    curve,
    certificate,
    protocol,
    provider,
    other_provider,
    early_data_size,
):
    ticket_file = str(tmp_path / TICKET_FILE)
    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    get_early_data_bytes(early_data_file, early_data_size)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        data_to_send=DATA_TO_SEND,
    )
    options.ticket_file = ticket_file
    options.early_data_file = early_data_file
    options.max_early_data = MAX_EARLY_DATA

    get_ticket_from_s2n_server(options, managed_process, provider, certificate)
    options.max_early_data = 0

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode

    s2n_server = managed_process(S2N, server_options, timeout=10)
    client = managed_process(provider, client_options, timeout=10)

    for results in client.get_results():
        results.assert_success()

    for results in s2n_server.get_results():
        results.assert_success()
        assert S2N_EARLY_DATA_MARKER not in results.stdout
        assert S2N_HRR_MARKER not in results.stdout
        assert to_bytes(S2N_EARLY_DATA_RECV_MARKER) not in results.stdout
        assert to_bytes(S2N_EARLY_DATA_REJECTED_MARKER) in results.stdout
        assert DATA_TO_SEND in results.stdout


"""
Test the S2N client attempting to send early data, but the server triggering a hello retry.

We trigger the HRR by configuring the server to only accept curves that the S2N client
does not send key shares for.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", S2N_HRR_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", SERVER_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize(
    "early_data_size",
    [int(MAX_EARLY_DATA / 2), int(MAX_EARLY_DATA - 1), MAX_EARLY_DATA, 1],
)
def test_s2n_client_with_early_data_rejected_via_hrr(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    curve,
    certificate,
    protocol,
    provider,
    other_provider,
    early_data_size,
):
    if provider == S2N:
        pytest.skip(
            "S2N does not respect ProviderOptions.curve, so does not trigger a retry"
        )

    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    early_data = get_early_data_bytes(early_data_file, early_data_size)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        reconnect=True,
    )
    options.ticket_file = None
    options.early_data_file = early_data_file
    options.max_early_data = MAX_EARLY_DATA

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode
    server_options.key = certificate.key  # Required for the initial connection
    server_options.cert = certificate.cert  # Required for the initial connection
    server_options.reconnects_before_exit = NUM_CONNECTIONS

    server = managed_process(provider, server_options, timeout=10)
    s2n_client = managed_process(S2N, client_options, timeout=10)

    for results in s2n_client.get_results():
        results.assert_success()
        assert S2N_EARLY_DATA_MARKER not in results.stdout
        assert S2N_HRR_MARKER in results.stdout
        assert (
            results.stdout.count(to_bytes(S2N_EARLY_DATA_REJECTED_MARKER))
            == NUM_RESUMES
        )

    for results in server.get_results():
        results.assert_success()
        assert early_data not in results.stdout


"""
Test the S2N server rejecting early data because of a hello retry request.

In order to trigger a successful retry, we need to force the peer to offer us a key share that
S2N doesn't support while still supporting at least one curve S2N does support.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", CLIENT_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize(
    "early_data_size",
    [int(MAX_EARLY_DATA / 2), int(MAX_EARLY_DATA - 1), MAX_EARLY_DATA, 1],
)
def test_s2n_server_with_early_data_rejected_via_hrr(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    curve,
    certificate,
    protocol,
    provider,
    other_provider,
    early_data_size,
):
    ticket_file = str(tmp_path / TICKET_FILE)
    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    early_data = get_early_data_bytes(early_data_file, early_data_size)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        curve=(S2N_UNSUPPORTED_CURVE + ":" + str(curve)),
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        data_to_send=DATA_TO_SEND,
    )
    options.ticket_file = ticket_file
    options.early_data_file = early_data_file
    options.max_early_data = MAX_EARLY_DATA

    get_ticket_from_s2n_server(options, managed_process, provider, certificate)

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode

    s2n_server = managed_process(S2N, server_options, timeout=10)
    client = managed_process(provider, client_options, timeout=10)

    for results in client.get_results():
        results.assert_success()
        assert early_data not in results.stdout

    for results in s2n_server.get_results():
        results.assert_success()
        assert S2N_EARLY_DATA_MARKER not in results.stdout
        assert S2N_HRR_MARKER in results.stdout
        assert to_bytes(S2N_EARLY_DATA_RECV_MARKER) not in results.stdout
        assert to_bytes(S2N_EARLY_DATA_REJECTED_MARKER) in results.stdout
        assert DATA_TO_SEND in results.stdout


"""
Test the S2N server fails if it receives too much early data.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", CLIENT_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("excess_early_data", [1, 10, MAX_EARLY_DATA])
def test_s2n_server_with_early_data_max_exceeded(
    managed_process,  # noqa: F811
    tmp_path,
    cipher,
    curve,
    certificate,
    protocol,
    provider,
    other_provider,
    excess_early_data,
):
    ticket_file = str(tmp_path / TICKET_FILE)
    early_data_file = str(tmp_path / EARLY_DATA_FILE)
    early_data = get_early_data_bytes(
        early_data_file, MAX_EARLY_DATA + excess_early_data
    )

    options = ProviderOptions(
        port=next(available_ports),
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        use_session_ticket=True,
        data_to_send=DATA_TO_SEND,
    )
    options.ticket_file = ticket_file
    options.early_data_file = early_data_file
    options.max_early_data = MAX_EARLY_DATA + excess_early_data

    get_ticket_from_s2n_server(options, managed_process, provider, certificate)
    options.max_early_data = MAX_EARLY_DATA

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode

    s2n_server = managed_process(S2N, server_options, timeout=10)
    client = managed_process(provider, client_options, timeout=10)

    for results in client.get_results():
        """
        We can't make any assertions about the client exit_code.
        To avoid blinding delays, s2nd doesn't call s2n_shutdown for a failed negotiation.
        That means that instead of sending close_notify, we just close the socket.
        Whether the peer interprets this as a failure or EoF depends on its state.
        """
        assert results.exception is None
        assert DATA_TO_SEND not in results.stdout

    for results in s2n_server.get_results():
        assert results.exception is None
        assert results.exit_code != 0
        # Full early data should not be reported
        assert early_data not in results.stdout
        # Partial early data should be reported
        assert (
            to_bytes(S2N_EARLY_DATA_RECV_MARKER) + early_data[:MAX_EARLY_DATA]
        ) in results.stdout
        assert to_bytes("Bad message encountered") in results.stderr



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
import pytest

from configuration import available_ports, ALL_TEST_CIPHERS, MINIMAL_TEST_CERTS
from common import ProviderOptions, Protocols, Signatures, data_bytes
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N, OpenSSL, GnuTLS
from utils import (
    invalid_test_parameters,
    get_parameter_name,
    get_expected_s2n_version,
    to_bytes,
)

all_sigs = [
    Signatures.RSA_SHA1,
    Signatures.RSA_SHA224,
    Signatures.RSA_SHA256,
    Signatures.RSA_SHA384,
    Signatures.RSA_SHA512,
    Signatures.ECDSA_SECP256r1_SHA256,
    Signatures.ECDSA_SECP384r1_SHA384,
    Signatures.ECDSA_SECP521r1_SHA512,
    Signatures.RSA_PSS_RSAE_SHA256,
    Signatures.RSA_PSS_PSS_SHA256,
    Signatures.ECDSA_SHA1,
    Signatures.ECDSA_SHA224,
    Signatures.ECDSA_SHA256,
    Signatures.ECDSA_SHA384,
    Signatures.ECDSA_SHA512,
]


def signature_marker(mode, protocol, cipher, signature):
    # In versions before TLS1.3, the server uses the negotiated signature scheme for the
    # ServerKeyExchange message. The server only sends the ServerKeyExchange message when using
    # a key exchange method that provides forward secrecy, ie, NOT static RSA.
    # So if using RSA key exchange, there is no actual "negotiated" signature scheme, because
    # the server never sends the client a signature scheme.
    #
    # This mostly has to be inferred from the RFCs, but this blog post is a pretty good summary
    # of the situation: https://timtaubert.de/blog/2016/07/the-evolution-of-signatures-in-tls/
    if mode == Provider.ServerMode and cipher.iana_standard_name.startswith(
        "TLS_RSA_WITH_"
    ):
        signature = Signatures.NONE
    elif protocol < Protocols.TLS12:
        # ECDSA by default hashes with SHA-1.
        #
        # This is inferred from extended version of TLS1.1 rfc- https://www.rfc-editor.org/rfc/rfc4492#section-5.10
        if signature.sig_type == "ECDSA":
            signature = Signatures.ECDSA_SHA1
        else:
            signature = Signatures.RSA_MD5_SHA1

    return to_bytes(
        "{mode} signature negotiated: {type}+{digest}".format(
            mode=mode.title(), type=signature.sig_type, digest=signature.sig_digest
        )
    )


def skip_ciphers(*args, **kwargs):
    provider = kwargs.get("provider")
    cert = kwargs.get("certificate")
    cipher = kwargs.get("cipher")
    protocol = kwargs.get("protocol")
    sigalg = kwargs.get("signature")

    if not provider.supports_signature(sigalg):
        return True

    if not cert.compatible_with_cipher(cipher):
        return True

    if not cert.compatible_with_sigalg(sigalg):
        return True

    if protocol > sigalg.max_protocol:
        return True

    if protocol < sigalg.min_protocol:
        return True

    return invalid_test_parameters(*args, **kwargs)


@pytest.mark.uncollect_if(func=skip_ciphers)
@pytest.mark.parametrize("cipher", ALL_TEST_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL, GnuTLS])
@pytest.mark.parametrize("other_provider", [S2N])
@pytest.mark.parametrize(
    "protocol",
    [Protocols.TLS13, Protocols.TLS12, Protocols.TLS11],
    ids=get_parameter_name,
)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("signature", all_sigs, ids=get_parameter_name)
@pytest.mark.parametrize(
    "client_auth",
    [True, False],
    ids=lambda val: "client-auth" if val else "no-client-auth",
)
def test_s2n_server_signature_algorithms(
    managed_process,  # noqa: F811
    cipher,
    provider,
    other_provider,
    protocol,
    certificate,
    signature,
    client_auth,
):
    port = next(available_ports)

    random_bytes = data_bytes(64)
    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        data_to_send=random_bytes,
        insecure=False,
        use_client_auth=client_auth,
        key=certificate.key,
        cert=certificate.cert,
        signature_algorithm=signature,
        protocol=protocol,
    )

    if provider == GnuTLS:
        # GnuTLS fails the CA verification. It must be run with this check disabled.
        client_options.extra_flags = ["--no-ca-verification"]

    server_options = copy.copy(client_options)
    server_options.extra_flags = None
    server_options.data_to_send = None
    server_options.mode = Provider.ServerMode
    server_options.key = certificate.key
    server_options.cert = certificate.cert

    server = managed_process(S2N, server_options, timeout=5)
    client = managed_process(provider, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()

    expected_version = get_expected_s2n_version(protocol, provider)

    for results in server.get_results():
        results.assert_success()
        assert (
            to_bytes("Actual protocol version: {}".format(expected_version))
            in results.stdout
        )
        assert (
            signature_marker(Provider.ServerMode, protocol, cipher, signature)
            in results.stdout
        )
        assert (
            signature_marker(Provider.ClientMode, protocol, cipher, signature)
            in results.stdout
        ) == client_auth
        assert random_bytes in results.stdout


@pytest.mark.flaky(reruns=5, reruns_delay=2)
@pytest.mark.uncollect_if(func=skip_ciphers)
@pytest.mark.parametrize("cipher", ALL_TEST_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL, GnuTLS])
@pytest.mark.parametrize("other_provider", [S2N])
@pytest.mark.parametrize(
    "protocol",
    [Protocols.TLS13, Protocols.TLS12, Protocols.TLS11],
    ids=get_parameter_name,
)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("signature", all_sigs, ids=get_parameter_name)
@pytest.mark.parametrize(
    "client_auth",
    [True, False],
    ids=lambda val: "client-auth" if val else "no-client-auth",
)
def test_s2n_client_signature_algorithms(
    managed_process,  # noqa: F811
    cipher,
    provider,
    other_provider,
    protocol,
    certificate,
    signature,
    client_auth,
):
    port = next(available_ports)

    random_bytes = data_bytes(64)
    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        data_to_send=random_bytes,
        insecure=True,
        use_client_auth=client_auth,
        key=certificate.key,
        cert=certificate.cert,
        protocol=protocol,
    )

    server_options = copy.copy(client_options)
    server_options.data_to_send = None
    server_options.mode = Provider.ServerMode
    server_options.key = certificate.key
    server_options.cert = certificate.cert
    server_options.trust_store = certificate.cert
    server_options.signature_algorithm = signature

    kill_marker = None
    if provider == GnuTLS:
        kill_marker = random_bytes

    server = managed_process(
        provider, server_options, timeout=5, kill_marker=kill_marker
    )
    client = managed_process(S2N, client_options, timeout=5)

    for results in server.get_results():
        results.assert_success()
        assert any([random_bytes in stream for stream in results.output_streams()])

    expected_version = get_expected_s2n_version(protocol, provider)

    for results in client.get_results():
        results.assert_success()
        assert (
            to_bytes("Actual protocol version: {}".format(expected_version))
            in results.stdout
        )
        assert (
            signature_marker(Provider.ServerMode, protocol, cipher, signature)
            in results.stdout
        )
        assert (
            signature_marker(Provider.ClientMode, protocol, cipher, signature)
            in results.stdout
        ) == client_auth



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
import pytest

from configuration import available_ports, PROTOCOLS
from common import ProviderOptions, Ciphers, Certificates, data_bytes
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N, OpenSSL, GnuTLS
from utils import (
    invalid_test_parameters,
    get_parameter_name,
    get_expected_s2n_version,
    to_bytes,
)


CIPHERS_TO_TEST = [
    Ciphers.AES256_SHA,
    Ciphers.ECDHE_ECDSA_AES256_SHA,
    Ciphers.AES256_GCM_SHA384,
]

CERTIFICATES_TO_TEST = [Certificates.RSA_4096_SHA384, Certificates.ECDSA_384]


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", CIPHERS_TO_TEST, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL, GnuTLS], ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("protocol", PROTOCOLS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", CERTIFICATES_TO_TEST, ids=get_parameter_name)
def test_s2n_server_low_latency(
    managed_process,  # noqa: F811
    cipher,
    provider,
    other_provider,
    protocol,
    certificate,
):
    if provider is OpenSSL and "openssl-1.0.2" in provider.get_version():
        pytest.skip(
            "{} does not allow setting max fragmentation for packets".format(provider)
        )

    port = next(available_ports)

    random_bytes = data_bytes(65519)
    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        data_to_send=random_bytes,
        insecure=True,
        protocol=protocol,
    )

    server_options = copy.copy(client_options)
    server_options.data_to_send = None
    server_options.mode = Provider.ServerMode
    server_options.extra_flags = ["--prefer-low-latency"]
    server_options.key = certificate.key
    server_options.cert = certificate.cert
    server_options.cipher = None

    server = managed_process(S2N, server_options, timeout=5)
    client = managed_process(provider, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()

    expected_version = get_expected_s2n_version(protocol, provider)

    for results in server.get_results():
        results.assert_success()
        assert (
            to_bytes("Actual protocol version: {}".format(expected_version))
            in results.stdout
        )
        assert random_bytes in results.stdout


def invalid_test_parameters_frag_len(*args, **kwargs):
    provider = kwargs.get("provider")
    frag_len = kwargs.get("frag_len")

    # Check to make sure frag_len is compatible with gnutls.
    if provider == GnuTLS:
        if frag_len > 4096:
            return True

    return invalid_test_parameters(*args, **kwargs)


@pytest.mark.uncollect_if(func=invalid_test_parameters_frag_len)
@pytest.mark.parametrize("cipher", CIPHERS_TO_TEST, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL, GnuTLS], ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("protocol", PROTOCOLS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", CERTIFICATES_TO_TEST, ids=get_parameter_name)
@pytest.mark.parametrize(
    "frag_len", [512, 2048, 8192, 12345, 16384], ids=get_parameter_name
)
def test_s2n_server_framented_data(
    managed_process,  # noqa: F811
    cipher,
    provider,
    other_provider,
    protocol,
    certificate,
    frag_len,
):
    if provider is OpenSSL and "openssl-1.0.2" in provider.get_version():
        pytest.skip(
            "{} does not allow setting max fragmentation for packets".format(provider)
        )

    port = next(available_ports)

    random_bytes = data_bytes(65519)
    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        data_to_send=random_bytes,
        insecure=True,
        record_size=frag_len,
        protocol=protocol,
    )

    server_options = copy.copy(client_options)
    server_options.extra_flags = None
    server_options.data_to_send = None
    server_options.mode = Provider.ServerMode
    server_options.key = certificate.key
    server_options.cert = certificate.cert
    server_options.cipher = None

    server = managed_process(S2N, server_options, timeout=5)
    client = managed_process(provider, client_options, timeout=5)

    for client_results in client.get_results():
        client_results.assert_success()

    expected_version = get_expected_s2n_version(protocol, provider)

    for server_results in server.get_results():
        server_results.assert_success()
        assert (
            to_bytes("Actual protocol version: {}".format(expected_version))
            in server_results.stdout
        )

        if provider == GnuTLS:
            # GnuTLS ignores data sent through stdin past frag_len up to the application data
            # packet length of 4096. so, just check to make sure data up to frag_len was received.
            assert random_bytes[:frag_len] in server_results.stdout
        else:
            assert random_bytes in server_results.stdout



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
import pytest
import subprocess

from processes import ManagedProcess
from providers import Provider, S2N

from common import ProviderOptions
from conftest import PATH_CONFIGURATION_KEY


@pytest.fixture
def managed_process(request: pytest.FixtureRequest):
    """
    Generic process manager. This could be used to launch any process as a background
    task and cleanup when finished.

    The reason a fixture is used, instead of creating a ManagedProcess() directly
    from the test, is to control the life of the process. Using the fixture
    allows cleanup after a test, even if a failure occurred.
    """
    processes = []
    # Indicates whether a launch was aborted. If so, non-graceful shutdown is allowed
    aborted = False

    def _fn(
        provider_class: Provider,
        options: ProviderOptions,
        timeout=5,
        send_marker=None,
        close_marker=None,
        expect_stderr=None,
        kill_marker=None,
        send_with_newline=None,
    ):
        best_effort_mode = request.config.getoption("--best-effort-NOT-FOR-CI")
        if best_effort_mode:
            # modify the `aborted` field in the generator object
            nonlocal aborted
            available_providers = request.config.stash[PATH_CONFIGURATION_KEY]
            if provider_class not in available_providers:
                aborted = True
                pytest.skip(f"{provider_class} is not available")

        provider = provider_class(options)
        cmd_line = provider.get_cmd_line()

        if (
            best_effort_mode
            and provider_class is S2N
            and not (cmd_line[0] == "s2nc" or cmd_line[0] == "s2nd")
        ):
            aborted = True
            pytest.skip("s2nc_head or s2nd_head not supported for best-effort")

        # The process will default to send markers in the providers.py file
        # if not specified by a test.
        if send_marker is not None:
            provider.ready_to_send_input_marker = send_marker
        if expect_stderr is None:
            expect_stderr = provider.expect_stderr
        if send_with_newline is None:
            send_with_newline = provider.send_with_newline
        p = ManagedProcess(
            cmd_line,
            provider.set_provider_ready,
            name=provider.get_name(cmd_line),
            wait_for_marker=provider.ready_to_test_marker,
            send_marker_list=provider.ready_to_send_input_marker,
            close_marker=close_marker,
            data_source=options.data_to_send,
            timeout=timeout,
            env_overrides=options.env_overrides,
            expect_stderr=expect_stderr,
            kill_marker=kill_marker,
            send_with_newline=send_with_newline,
        )

        processes.append(p)
        with p.ready_condition:
            p.start()
            with provider._provider_ready_condition:
                # Don't continue processing until the provider has indicated it is ready.
                provider._provider_ready_condition.wait_for(
                    provider.is_provider_ready, timeout
                )
        return p

    try:
        yield _fn
    except Exception as _:
        # The ManagedProcess already prints information to stdout, so there
        # is nothing to capture here.
        pass
    finally:
        for p in processes:
            # Always print the results
            if p.results:
                width = 90
                padchar = "#"

                print(padchar * width)
                print(f"  {p.cmd_line[0]}  ".center(width, padchar))
                print(padchar * width)

                print(f"Command line:\n\t{' '.join(p.cmd_line)}")
                print(f"Exit code:\n\t {p.results.exit_code}")
                print("")

                print("  Stdout  ".center(width, padchar))
                print(p.results.stdout.decode("utf-8", "backslashreplace"))
                print("")

                print("  Stderr  ".center(width, padchar))
                print(p.results.stderr.decode("utf-8", "backslashreplace"))
                print("")

            # Whether the processes succeeded or not, clean them up.
            if aborted:
                p.kill()
            else:
                p.join()


def _swap_mtu(device, new_mtu):
    """
    Swap the device's current MTU for the requested MTU.
    Return the original MTU so it can be reset later.
    """
    cmd = ["ip", "link", "show", device]
    p = subprocess.Popen(
        cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    mtu = 65536
    for line in p.stdout.readlines():
        s = line.decode("utf-8")
        pieces = s.split(" ")
        if len(pieces) >= 4 and pieces[3] == "mtu":
            mtu = int(pieces[4])

    p.wait()

    subprocess.call(["ip", "link", "set", device, "mtu", str(new_mtu)])

    return int(mtu)


@pytest.fixture(scope="module")
def custom_mtu():
    """
    This fixture will swap the loopback's MTU from the default
    to 1500, which is more reasonable for a network device.
    Using a fixture allows us to reset the MTU even if the test
    fails.

    These values are all hardcoded because they are only used
    from a single test. This simplifies the use of the fixture.
    """
    if os.geteuid() != 0:
        pytest.skip("Test needs root privileges to modify lo MTU")

    original_mtu = _swap_mtu("lo", 1500)
    yield
    _swap_mtu("lo", original_mtu)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest

from configuration import (
    available_ports,
    PROTOCOLS,
    MINIMAL_TEST_CIPHERS,
    MINIMAL_TEST_CERTS,
)
from common import ProviderOptions, data_bytes
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N, OpenSSL, GnuTLS
from utils import invalid_test_parameters, get_parameter_name, to_bytes, to_string

SEND_DATA_SIZE = 2**14

# CLOSE_MARKER must a substring of SEND_DATA exactly once, and must be its suffix
CLOSE_MARKER = "unique-suffix-close-marker"

SEND_DATA = data_bytes(SEND_DATA_SIZE - len(CLOSE_MARKER)) + to_bytes(CLOSE_MARKER)
SEND_DATA_STRING = to_string(SEND_DATA)

K_BYTES = 1024
SEND_BUFFER_SIZE_MIN = 1034
SEND_BUFFER_SIZE_MIN_RECOMMENDED = 2 * K_BYTES
SEND_BUFFER_SIZE_MULTI_RECORD = 17 * K_BYTES
SEND_BUFFER_SIZE_PREFER_THROUGHPUT = 35 * K_BYTES
SEND_BUFFER_SIZE_HUGE = 512 * K_BYTES

SEND_BUFFER_SIZES = [
    SEND_BUFFER_SIZE_MIN,
    SEND_BUFFER_SIZE_MIN_RECOMMENDED,
    SEND_BUFFER_SIZE_MULTI_RECORD,
    SEND_BUFFER_SIZE_PREFER_THROUGHPUT,
    SEND_BUFFER_SIZE_HUGE,
]

FRAGMENT_PREFERENCE = [None, "--prefer-low-latency", "--prefer-throughput"]


def test_SEND_BUFFER_SIZE_MIN_is_s2ns_min_buffer_size(managed_process):  # noqa: F811
    port = next(available_ports)

    s2n_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        data_to_send="test",
        extra_flags=["--buffered-send", SEND_BUFFER_SIZE_MIN],
    )

    s2nd = managed_process(S2N, s2n_options)

    s2n_options.mode = Provider.ClientMode
    s2n_options.extra_flags = ["--buffered-send", SEND_BUFFER_SIZE_MIN - 1]
    s2nc = managed_process(S2N, s2n_options)

    for results in s2nc.get_results():
        assert "Error setting send buffer size" in str(results.stderr)
        assert results.exit_code != 0

    for results in s2nd.get_results():
        assert results.exit_code != 0


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", MINIMAL_TEST_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [GnuTLS, OpenSSL, S2N], ids=get_parameter_name)
@pytest.mark.parametrize("protocol", PROTOCOLS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("buffer_size", SEND_BUFFER_SIZES, ids=get_parameter_name)
@pytest.mark.parametrize(
    "fragment_preference", FRAGMENT_PREFERENCE, ids=get_parameter_name
)
def test_s2n_server_buffered_send(
    managed_process,  # noqa: F811
    cipher,
    provider,
    protocol,
    certificate,
    buffer_size,
    fragment_preference,
):
    # Communication Timeline
    # Client [S2N|OpenSSL|GnuTLS]  | Server [S2N]
    # Handshake                    | Handshake
    #                              | Send SEND_DATA (with CLOSE_MARKER)
    # Closes with CLOSE_MARKER     | Close
    port = next(available_ports)

    provider_client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        data_to_send=None,
        insecure=True,
        protocol=protocol,
        verbose=False,
    )

    extra_flags = ["--buffered-send", buffer_size]
    if fragment_preference is not None:
        extra_flags.append(fragment_preference)

    s2n_server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        cipher=cipher,
        data_to_send=SEND_DATA,
        insecure=True,
        protocol=protocol,
        key=certificate.key,
        cert=certificate.cert,
        extra_flags=extra_flags,
    )

    server = managed_process(
        S2N, s2n_server_options, send_marker=[S2N.get_send_marker()]
    )
    client = managed_process(
        provider, provider_client_options, close_marker=CLOSE_MARKER
    )

    for results in client.get_results():
        assert SEND_DATA_STRING in str(results.stdout)
        results.assert_success()

    for results in server.get_results():
        # the server should close without error
        # but there is otherwise nothing of interest on stdout
        results.assert_success()


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", MINIMAL_TEST_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [S2N, OpenSSL], ids=get_parameter_name)
@pytest.mark.parametrize("protocol", PROTOCOLS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", MINIMAL_TEST_CERTS, ids=get_parameter_name)
@pytest.mark.parametrize("buffer_size", SEND_BUFFER_SIZES, ids=get_parameter_name)
@pytest.mark.parametrize(
    "fragment_preference", FRAGMENT_PREFERENCE, ids=get_parameter_name
)
def test_s2n_client_buffered_send(
    managed_process,  # noqa: F811
    cipher,
    provider,
    protocol,
    certificate,
    buffer_size,
    fragment_preference,
):
    # Communication Timeline
    # Client [S2N]                       | Server [S2N|OpenSSL]
    # Handshake                          | Handshake
    # Send SEND_DATA (with CLOSE_MARKER) | Receive the Data Bytes
    # Close                              | Close on CLOSE_MARKER
    port = next(available_ports)

    extra_flags = ["--buffered-send", buffer_size]
    if fragment_preference is not None:
        extra_flags.append(fragment_preference)

    s2n_client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        data_to_send=SEND_DATA,
        insecure=True,
        protocol=protocol,
        extra_flags=extra_flags,
    )

    provider_server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        cipher=cipher,
        insecure=True,
        protocol=protocol,
        key=certificate.key,
        cert=certificate.cert,
        verbose=False,
    )

    server = managed_process(
        provider, provider_server_options, close_marker=CLOSE_MARKER
    )
    client = managed_process(S2N, s2n_client_options)

    for results in client.get_results():
        results.assert_success()

    for results in server.get_results():
        assert SEND_DATA_STRING in str(results.stdout)
        results.assert_success()



#!/usr/bin/env python3

import os
import re
import subprocess

tag = subprocess.run(['git', 'describe', '--tags'],
                     capture_output=True, check=True,
                     text=True).stdout.strip()
# convert v0.2.12-2-g50254a9 to 0.2.12
# test-version-exists will ensure to not include non-tagged commits
version = tag.strip("v").split('-', 1)[0]

init_path = os.path.join(os.path.dirname(__file__), '..', 'awscrt', '__init__.py')
print("Updating awscrt.__version__ to version {}".format(version))
contents = None
with open(init_path, 'r+') as init_py:
    contents = init_py.read()

contents = re.sub(r"__version__ = '[^']+'", f"__version__ = '{version}'", contents)

with open(init_path, 'w') as init_py:
    init_py.write(contents)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest

from configuration import available_ports, ALL_TEST_CIPHERS, ALL_TEST_CURVES, PROTOCOLS
from common import ProviderOptions, data_bytes, Certificates
from fixtures import managed_process  # noqa: F401
from constants import TEST_OCSP_DIRECTORY
from providers import Provider, S2N, OpenSSL, GnuTLS
from utils import invalid_test_parameters, get_parameter_name
from global_flags import get_flag, S2N_PROVIDER_VERSION


OCSP_CERTS = [Certificates.OCSP, Certificates.OCSP_ECDSA]


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", ALL_TEST_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [S2N, OpenSSL, GnuTLS], ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", PROTOCOLS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", OCSP_CERTS, ids=get_parameter_name)
def test_s2n_client_ocsp_response(
    managed_process,  # noqa: F811
    cipher,
    provider,
    curve,
    protocol,
    certificate,
):
    if "boringssl" in get_flag(S2N_PROVIDER_VERSION):
        pytest.skip("s2n-tls client with boringssl does not support ocsp")

    port = next(available_ports)

    random_bytes = data_bytes(128)
    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        data_to_send=random_bytes,
        enable_client_ocsp=True,
    )

    server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        key=certificate.key,
        cert=certificate.cert,
        ocsp_response={
            "RSA": TEST_OCSP_DIRECTORY + "ocsp_response.der",
            "EC": TEST_OCSP_DIRECTORY + "ocsp_ecdsa_response.der",
        }.get(certificate.algorithm),
    )

    kill_marker = None
    if provider == GnuTLS:
        # The gnutls-serv process will never exit on its own, so should be killed
        # to avoid a long timeout. However, we must NOT kill it until it sends
        # the close_notify that the s2n-tls client expects. The only good signal
        # for this is a debug message indicating that the alert was sent.
        #
        # The full debug message is something like:
        # "Sent Packet[4] Alert(21) in epoch 2 and length: 24"
        # but the packet number and epoch can vary. We are therefore forced to
        # only match on a very narrow substring, which could prove brittle.
        kill_marker = b"Alert(21) in epoch"
        server_options.extra_flags = ["-d", "5"]

    server = managed_process(
        provider, server_options, timeout=30, kill_marker=kill_marker
    )
    client = managed_process(S2N, client_options, timeout=30)

    for client_results in client.get_results():
        client_results.assert_success()
        assert b"OCSP response received" in client_results.stdout

    for server_results in server.get_results():
        server_results.assert_success()
        # Avoid debugging information that sometimes gets inserted after the first character.
        assert (
            random_bytes[1:] in server_results.stdout
            or random_bytes[1:] in server_results.stderr
        )


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", ALL_TEST_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [GnuTLS, OpenSSL], ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", PROTOCOLS, ids=get_parameter_name)
@pytest.mark.parametrize("certificate", OCSP_CERTS, ids=get_parameter_name)
def test_s2n_server_ocsp_response(
    managed_process,  # noqa: F811
    cipher,
    provider,
    curve,
    protocol,
    certificate,
):
    port = next(available_ports)

    random_bytes = data_bytes(128)
    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        data_to_send=random_bytes,
        enable_client_ocsp=True,
    )

    server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        cipher=cipher,
        curve=curve,
        protocol=protocol,
        insecure=True,
        key=certificate.key,
        cert=certificate.cert,
        ocsp_response={
            "RSA": TEST_OCSP_DIRECTORY + "ocsp_response.der",
            "EC": TEST_OCSP_DIRECTORY + "ocsp_ecdsa_response.der",
        }.get(certificate.algorithm),
    )

    server = managed_process(S2N, server_options, timeout=90)
    client = managed_process(provider, client_options, timeout=90)

    for client_results in client.get_results():
        client_results.assert_success()
        assert any(
            [
                {
                    GnuTLS: b"OCSP Response Information:\n\tResponse Status: Successful",
                    OpenSSL: b"OCSP Response Status: successful",
                }.get(provider)
                in stream
                for stream in client_results.output_streams()
            ]
        )

    for server_results in server.get_results():
        server_results.assert_success()
        # Avoid debugging information that sometimes gets inserted after the first character.
        assert any(
            [random_bytes[1:] in stream for stream in server_results.output_streams()]
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.


from test import NativeResourceTest
from awscrt import checksums
import unittest
import sys


class TestChecksums(NativeResourceTest):

    def test_crc32_zeros_one_shot(self):
        output = checksums.crc32(bytes(32))
        expected = 0x190A55AD
        self.assertEqual(expected, output)

    def test_crc32_zeros_iterated(self):
        output = 0
        for i in range(32):
            output = checksums.crc32(bytes(1), output)
        expected = 0x190A55AD
        self.assertEqual(expected, output)

    def test_crc32_values_one_shot(self):
        output = checksums.crc32(''.join(chr(i) for i in range(32)))
        expected = 0x91267E8A
        self.assertEqual(expected, output)

    def test_crc32_values_iterated(self):
        output = 0
        for i in range(32):
            output = checksums.crc32(chr(i), output)
        expected = 0x91267E8A
        self.assertEqual(expected, output)

    def test_crc32_large_buffer(self):
        # stress test gil optimization for 32 bit architecture which cannot handle huge buffer
        large_buffer = bytes(25 * 2**20)
        val = checksums.crc32(large_buffer)
        self.assertEqual(0x72103906, val)

    def test_crc32_huge_buffer(self):
        # stress the internal logic that handles buffers larger than C's INT_MAX
        if sys.platform.startswith('freebsd'):
            # Skip this test for freebsd, as it simply crashes instead of raising exception in this case
            raise unittest.SkipTest('Skip this test for freebsd')
        try:
            INT_MAX = 2**32 - 1
            huge_buffer = bytes(INT_MAX + 5)
        except (MemoryError, OverflowError):
            raise unittest.SkipTest('Machine cant allocate giant buffer for giant buffer test')
        val = checksums.crc32(huge_buffer)
        self.assertEqual(0xc622f71d, val)

    def test_crc32c_zeros_one_shot(self):
        output = checksums.crc32c(bytes(32))
        expected = 0x8A9136AA
        self.assertEqual(expected, output)

    def test_crc32c_zeros_iterated(self):
        output = 0
        for i in range(32):
            output = checksums.crc32c(bytes(1), output)
        expected = 0x8A9136AA
        self.assertEqual(expected, output)

    def test_crc32c_values_one_shot(self):
        output = checksums.crc32c(''.join(chr(i) for i in range(32)))
        expected = 0x46DD794E
        self.assertEqual(expected, output)

    def test_crc32c_values_iterated(self):
        output = 0
        for i in range(32):
            output = checksums.crc32c(chr(i), output)
        expected = 0x46DD794E
        self.assertEqual(expected, output)

    def test_crc32c_large_buffer(self):
        # stress test gil optimization for 32 bit architecture which cannot handle huge buffer
        large_buffer = bytes(25 * 2**20)
        val = checksums.crc32c(large_buffer)
        self.assertEqual(0xfb5b991d, val)

    def test_crc32c_huge_buffer(self):
        if sys.platform.startswith('freebsd'):
            # Skip this test for freebsd, as it simply crashes instead of raising exception in this case
            raise unittest.SkipTest('Skip this test for freebsd')
        try:
            INT_MAX = 2**32 - 1
            huge_buffer = bytes(INT_MAX + 5)
        except BaseException:
            raise unittest.SkipTest('Machine cant allocate giant buffer for giant buffer test')
        val = checksums.crc32c(huge_buffer)
        self.assertEqual(0x572a7c8a, val)

    def test_crc64nvme_zeros_one_shot(self):
        output = checksums.crc64nvme(bytes(32))
        expected = 0xcf3473434d4ecf3b
        self.assertEqual(expected, output)

    def test_crc64nvme_zeros_iterated(self):
        output = 0
        for i in range(32):
            output = checksums.crc64nvme(bytes(1), output)
        expected = 0xcf3473434d4ecf3b
        self.assertEqual(expected, output)

    def test_crc64nvme_values_one_shot(self):
        output = checksums.crc64nvme(''.join(chr(i) for i in range(32)))
        expected = 0xb9d9d4a8492cbd7f
        self.assertEqual(expected, output)

    def test_crc64nvme_values_iterated(self):
        output = 0
        for i in range(32):
            output = checksums.crc64nvme(chr(i), output)
        expected = 0xb9d9d4a8492cbd7f
        self.assertEqual(expected, output)

    def test_crc64nvme_large_buffer(self):
        # stress test gil optimization for 32 bit architecture which cannot handle huge buffer
        large_buffer = bytes(25 * 2**20)
        val = checksums.crc64nvme(large_buffer)
        self.assertEqual(0x5b6f5045463ca45e, val)

    def test_crc64nvme_huge_buffer(self):
        if sys.platform.startswith('freebsd'):
            # Skip this test for freebsd, as it simply crashes instead of raising exception in this case
            raise unittest.SkipTest('Skip this test for freebsd')
        try:
            INT_MAX = 2**32 - 1
            huge_buffer = bytes(INT_MAX + 5)
        except BaseException:
            raise unittest.SkipTest('Machine cant allocate giant buffer for giant buffer test')
        val = checksums.crc64nvme(huge_buffer)
        self.assertEqual(0x2645c28052b1fbb0, val)


if __name__ == '__main__':
    unittest.main()




import Builder
import glob, os, sys

class ClangTidy(Builder.Action):
    def run(self, env):
        sh = env.shell
        clang_tidy = env.find_llvm_tool('clang-tidy')[0]
        if not clang_tidy:
            print("No clang-tidy executable could be found, installing...")
            sh.exec("sudo", "apt", "install", "-y", "clang-tidy-9")
            clang_tidy = env.find_llvm_tool('clang-tidy')[0]
            if not clang_tidy:
                print("No clang-tidy executable could be found")
                sys.exit(1)

        source_dir = sh.cwd()
        build_dir = os.path.join(source_dir, 'build')
        sources = [os.path.join(source_dir, file) for file in glob.glob(
            'source/**/*.c') if not ('windows' in file or 'android' in file)]

        return [
            Builder.DownloadDependencies(),
            Builder.CMakeBuild(),
            Builder.Script([
                [clang_tidy, '-p', build_dir] + sources
            ])
        ]
        



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
import subprocess
import pytest
import threading

from common import (
    ProviderOptions,
    Certificates,
    Ciphers,
    Curves,
    Protocols,
    Signatures,
    Cert,
)
from global_flags import get_flag, S2N_PROVIDER_VERSION, S2N_FIPS_MODE
from stat import S_IMODE


class Provider(object):
    """
    A provider defines a specific provider of TLS. This could be
    S2N, OpenSSL, BoringSSL, etc.
    """

    ClientMode = "client"
    ServerMode = "server"

    def __init__(self, options: ProviderOptions):
        # If the provider includes stderr output on a success, set this to True.
        self.expect_stderr = False

        # If the test should wait for a specific output message before beginning,
        # put that message in ready_to_test_marker
        self.ready_to_test_marker = None

        # If a newline character should be added to messages being sent. Required
        # with some providers to properly write to stdin.
        self.send_with_newline = False

        # By default, we expect clients to send, but not servers.
        if options.mode == Provider.ClientMode:
            self.ready_to_send_input_marker = self.get_send_marker()
        else:
            self.ready_to_send_input_marker = None

        # Allows users to determine if the provider is ready to begin testing
        self._provider_ready_condition = threading.Condition()
        self._provider_ready = False

        if type(options) is not ProviderOptions:
            raise TypeError

        self.options = options
        if self.options.mode == Provider.ServerMode:
            self.cmd_line = self.setup_server()  # lgtm [py/init-calls-subclass]
        elif self.options.mode == Provider.ClientMode:
            self.cmd_line = self.setup_client()  # lgtm [py/init-calls-subclass]

    def setup_client(self):
        """
        Provider specific setup code goes here.
        This will probably include creating the command line based on ProviderOptions.
        """
        raise NotImplementedError

    def setup_server(self):
        """
        Provider specific setup code goes here.
        This will probably include creating the command line based on ProviderOptions.
        """
        raise NotImplementedError

    @classmethod
    def get_send_marker(cls):
        """
        This should be the last message printed before the client/server can send data.
        """
        return None

    @classmethod
    def supports_protocol(cls, protocol):
        raise NotImplementedError

    @classmethod
    def supports_cipher(cls, cipher, with_curve=None):
        raise NotImplementedError

    @classmethod
    def supports_signature(cls, signature):
        return True

    def get_cmd_line(self):
        return self.cmd_line

    def is_provider_ready(self):
        return self._provider_ready is True

    def set_provider_ready(self):
        with self._provider_ready_condition:
            self._provider_ready = True
            self._provider_ready_condition.notify()

    @classmethod
    def supports_certificate(cls, cert: Cert):
        return True

    @classmethod
    def get_name(cls, cmd_line):
        return cmd_line[0]


class Tcpdump(Provider):
    """
    TcpDump is used by the dynamic record test. It only needs to watch
    a handful of packets before it can exit.

    This class still follows the provider setup, but all values are hardcoded
    because this isn't expected to be used outside of the dynamic record test.
    """

    def __init__(self, options: ProviderOptions):
        Provider.__init__(self, options)

    def setup_client(self):
        self.ready_to_test_marker = "listening on lo"
        tcpdump_filter = "dst port {}".format(self.options.port)

        cmd_line = [
            "tcpdump",
            # Line buffer the output
            "-l",
            # Only read 10 packets before exiting. This is enough to find a large
            # packet, and still exit before the timeout.
            "-c",
            "10",
            # Watch the loopback device
            "-i",
            "lo",
            # Don't resolve IP addresses
            "-nn",
            # Set the buffer size to 1k
            "-B",
            "1024",
            tcpdump_filter,
        ]

        return cmd_line


class S2N(Provider):
    """
    The S2N provider translates flags into s2nc/s2nd command line arguments.
    """

    def __init__(self, options: ProviderOptions):
        Provider.__init__(self, options)

        self.send_with_newline = True  # lgtm [py/overwritten-inherited-attribute]

    @classmethod
    def get_send_marker(cls):
        return "s2n is ready"

    @classmethod
    def _pss_supported(cls):
        # RSA-PSS is unsupported for openssl-1.0
        # libressl and boringssl are disabled because of configuration issues
        # see https://github.com/aws/s2n-tls/issues/3250
        PSS_UNSUPPORTED_LIBCRYPTOS = {"libressl", "boringssl", "openssl-1.0"}
        for libcrypto in PSS_UNSUPPORTED_LIBCRYPTOS:
            # e.g. "openssl-1.0" in "openssl-1.0.2-fips"
            if libcrypto in get_flag(S2N_PROVIDER_VERSION):
                return False
        return True

    @classmethod
    def supports_certificate(cls, cert: Cert):
        if not cls._pss_supported() and cert.algorithm == "RSAPSS":
            return False
        # https://github.com/aws/s2n-tls/issues/5200
        if (
            "openssl-3.0-fips" in get_flag(S2N_PROVIDER_VERSION)
            and "RSA_1024" in cert.name
        ):
            return False
        return True

    @classmethod
    def supports_protocol(cls, protocol):
        if not cls._pss_supported() and protocol == Protocols.TLS13:
            return False

        # SSLv3 cannot be negotiated in FIPS mode with libcryptos other than AWS-LC.
        if all(
            [
                protocol == Protocols.SSLv3,
                get_flag(S2N_FIPS_MODE),
                "awslc" not in get_flag(S2N_PROVIDER_VERSION),
            ]
        ):
            return False

        return True

    @classmethod
    def supports_cipher(cls, cipher, with_curve=None):
        # Disable chacha20 and RC4 tests in libcryptos that don't support those
        # algorithms
        unsupported_configurations = {
            "CHACHA20": ["openssl-1.0.2", "libressl"],
            "RC4": ["openssl-3"],
        }

        for (
            unsupported_cipher,
            unsupported_libcryptos,
        ) in unsupported_configurations.items():
            # the queried cipher has some libcrypto's that don't support it
            # e.g. "RC4" in "TLS_ECDHE_RSA_WITH_RC4_128_SHA"
            if unsupported_cipher in cipher.name:
                current_libcrypto = get_flag(S2N_PROVIDER_VERSION)
                for lc in unsupported_libcryptos:
                    # e.g. "openssl-3" in "openssl-3.0.7"
                    if lc in current_libcrypto:
                        return False
        return True

    @classmethod
    def supports_signature(cls, signature):
        # Disable RSA_PSS_RSAE_SHA256 in unsupported libcryptos
        if (
            any(
                [
                    libcrypto in get_flag(S2N_PROVIDER_VERSION)
                    for libcrypto in ["openssl-1.0.2", "libressl", "boringssl"]
                ]
            )
            and signature == Signatures.RSA_PSS_RSAE_SHA256
        ):
            return False

        return True

    def setup_client(self):
        """
        Using the passed ProviderOptions, create a command line.
        """
        cmd_line = []
        if self.options.use_mainline_version is True:
            cmd_line.append("s2nc_head")
        else:
            cmd_line.append("s2nc")
        cmd_line.append("--non-blocking")

        # Tests requiring reconnects can't wait on echo data
        if self.options.echo and not self.options.reconnect:
            cmd_line.append("-e")

        if self.options.use_session_ticket is False:
            cmd_line.append("-T")

        if self.options.insecure is True:
            cmd_line.append("--insecure")
        elif self.options.trust_store:
            cmd_line.extend(["-f", self.options.trust_store])
        elif self.options.cert:
            cmd_line.extend(["-f", self.options.cert])

        if self.options.reconnect is True:
            cmd_line.append("-r")

        # If the test provided a cipher (security policy) that is compatible with
        # s2n, we'll use it. Otherwise, default to the appropriate `test_all` policy.
        cipher_prefs = "test_all_tls12"
        if self.options.protocol is Protocols.TLS13:
            cipher_prefs = "test_all"
        if self.options.cipher and self.options.cipher.s2n:
            cipher_prefs = self.options.cipher.name

        cmd_line.extend(["-c", cipher_prefs])

        if self.options.use_client_auth:
            if self.options.key:
                cmd_line.extend(["--key", self.options.key])
            if self.options.cert:
                cmd_line.extend(["--cert", self.options.cert])

        if get_flag(S2N_FIPS_MODE):
            cmd_line.append("--enter-fips-mode")

        if self.options.enable_client_ocsp:
            cmd_line.extend(["--status"])

        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        cmd_line.extend([self.options.host, self.options.port])

        # Clients are always ready to connect
        self.set_provider_ready()

        return cmd_line

    def setup_server(self):
        # s2nd prints this message after it begins listening for connections
        self.ready_to_test_marker = "Listening on"

        """
        Using the passed ProviderOptions, create a command line.
        """
        cmd_line = []
        if self.options.use_mainline_version is True:
            cmd_line.append("s2nd_head")
        else:
            cmd_line.append("s2nd")
        cmd_line.extend(["-X", "--self-service-blinding", "--non-blocking"])

        if self.options.key is not None:
            cmd_line.extend(["--key", self.options.key])
        if self.options.cert is not None:
            cmd_line.extend(["--cert", self.options.cert])

        if self.options.insecure is True:
            cmd_line.append("--insecure")
        elif self.options.trust_store:
            cmd_line.extend(["-t", self.options.trust_store])
        elif self.options.cert:
            cmd_line.extend(["-t", self.options.cert])

        # If the test provided a cipher (security policy) that is compatible with
        # s2n, we'll use it. Otherwise, default to the appropriate `test_all` policy.
        cipher_prefs = "test_all_tls12"
        if self.options.protocol is Protocols.TLS13:
            cipher_prefs = "test_all"
        if self.options.cipher and self.options.cipher.s2n:
            cipher_prefs = self.options.cipher.name

        cmd_line.extend(["-c", cipher_prefs])

        if not self.options.echo:
            cmd_line.append("-n")

        if self.options.use_client_auth is True:
            cmd_line.append("-m")

        if self.options.use_session_ticket is False:
            cmd_line.append("-T")

        if self.options.reconnects_before_exit is not None:
            cmd_line.append(
                "--max-conns={}".format(self.options.reconnects_before_exit)
            )

        if get_flag(S2N_FIPS_MODE):
            cmd_line.append("--enter-fips-mode")

        if self.options.ocsp_response is not None:
            cmd_line.extend(["--ocsp", self.options.ocsp_response])

        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        cmd_line.extend([self.options.host, self.options.port])

        return cmd_line


class OpenSSL(Provider):
    result = subprocess.run(
        ["openssl", "version"], shell=False, capture_output=True, text=True
    )
    # After splitting, version_str would be: ["OpenSSL", "3.0.8", "7", "Feb", "2023\n"]
    version_str = result.stdout.split(" ")
    # e.g., "OpenSSL"
    provider = version_str[0]
    # e.g., "3.0.8"
    version_openssl = version_str[1]

    def __init__(self, options: ProviderOptions):
        Provider.__init__(self, options)
        # We print some OpenSSL logging that includes stderr
        self.expect_stderr = True  # lgtm [py/overwritten-inherited-attribute]
        # Current provider needs 1.1.x https://github.com/aws/s2n-tls/issues/3963
        self.at_least_openssl_1_1()

    @classmethod
    def get_send_marker(cls):
        return "Verify return code"

    def _join_ciphers(self, ciphers):
        """
        Given a list of ciphers, join the names with a ':' like OpenSSL expects
        """
        assert type(ciphers) is list

        cipher_list = []
        for c in ciphers:
            cipher_list.append(c.name)

        ciphers = ":".join(cipher_list)

        return ciphers

    def _cipher_to_cmdline(self, cipher):
        cmdline = list()

        ciphers = []
        if type(cipher) is list:
            # In the case of a cipher list we need to be sure TLS13 specific ciphers aren't
            # mixed with ciphers from previous versions
            is_tls13_or_above = cipher[0].min_version >= Protocols.TLS13
            mismatch = [
                c
                for c in cipher
                if (c.min_version >= Protocols.TLS13) != is_tls13_or_above
            ]

            if len(mismatch) > 0:
                raise Exception(
                    "Cannot combine ciphers for TLS1.3 or above with older ciphers: {}".format(
                        [c.name for c in cipher]
                    )
                )

            ciphers.append(self._join_ciphers(cipher))
        else:
            is_tls13_or_above = cipher.min_version >= Protocols.TLS13
            ciphers.append(cipher.name)

        if is_tls13_or_above:
            cmdline.append("-ciphersuites")
        else:
            cmdline.append("-cipher")

        return cmdline + ciphers

    @classmethod
    def get_version(cls):
        return cls.version_openssl

    @classmethod
    def get_provider(cls):
        return cls.provider

    @classmethod
    def supports_protocol(cls, protocol):
        if OpenSSL.get_version()[0:3] == "1.1":
            return protocol not in (Protocols.SSLv3,)
        elif OpenSSL.get_version()[0:3] == "3.0":
            return protocol not in (Protocols.SSLv3, Protocols.TLS10, Protocols.TLS11)
        else:
            return True

    @classmethod
    def supports_certificate(cls, cert: Cert):
        if OpenSSL.get_version()[0:3] >= "3.0":
            return cert not in (
                Certificates.RSA_1024_SHA256,
                Certificates.RSA_1024_SHA384,
                Certificates.RSA_1024_SHA512,
            )

        return True

    @classmethod
    def supports_cipher(cls, cipher, with_curve=None):
        return True

    def at_least_openssl_1_1(self) -> None:
        if OpenSSL.get_version() < "1.1":
            raise FileNotFoundError(
                f"Openssl version returned {OpenSSL.get_version()}, expected at least 1.1.x."
            )

    def setup_client(self):
        cmd_line = ["openssl", "s_client"]
        cmd_line.extend(
            ["-connect", "{}:{}".format(self.options.host, self.options.port)]
        )

        # Additional debugging that will be captured incase of failure
        if self.options.verbose:
            cmd_line.append("-debug")

        cmd_line.extend(["-tlsextdebug", "-state"])

        if self.options.key is not None:
            cmd_line.extend(["-key", self.options.key])

        # Unlike s2n, OpenSSL allows us to be much more specific about which TLS
        # protocol to use.
        if self.options.protocol == Protocols.TLS13:
            cmd_line.append("-tls1_3")
        elif self.options.protocol == Protocols.TLS12:
            cmd_line.append("-tls1_2")
        elif self.options.protocol == Protocols.TLS11:
            cmd_line.append("-tls1_1")
        elif self.options.protocol == Protocols.TLS10:
            cmd_line.append("-tls1")
        elif self.options.protocol == Protocols.SSLv3:
            cmd_line.append("-ssl3")

        if self.options.cipher is not None:
            cmd_line.extend(self._cipher_to_cmdline(self.options.cipher))

        if self.options.curve is not None:
            cmd_line.extend(["-curves", str(self.options.curve)])

        if self.options.use_client_auth:
            if self.options.key:
                cmd_line.extend(["-key", self.options.key])
            if self.options.cert:
                cmd_line.extend(["-cert", self.options.cert])

        if self.options.reconnect is True:
            cmd_line.append("-reconnect")

        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        if self.options.server_name is not None:
            cmd_line.extend(["-servername", self.options.server_name])
            if self.options.verify_hostname is not None:
                cmd_line.extend(["-verify_hostname", self.options.server_name])

        if self.options.enable_client_ocsp:
            cmd_line.append("-status")

        if self.options.signature_algorithm is not None:
            cmd_line.extend(["-sigalgs", self.options.signature_algorithm.name])

        if self.options.record_size is not None:
            cmd_line.extend(["-max_send_frag", str(self.options.record_size)])

        # Clients are always ready to connect
        self.set_provider_ready()

        return cmd_line

    def setup_server(self):
        # s_server prints this message before it is ready to send/receive data
        self.ready_to_test_marker = "ACCEPT"

        cmd_line = ["openssl", "s_server"]
        cmd_line.extend(["-accept", "{}".format(self.options.port)])

        if self.options.reconnects_before_exit is not None:
            # If the user request a specific reconnection count, set it here
            cmd_line.extend(["-naccept", str(self.options.reconnects_before_exit)])
        else:
            # Exit after the first connection by default
            cmd_line.extend(["-naccept", "1"])

        # Additional debugging that will be captured incase of failure
        if self.options.verbose:
            cmd_line.append("-debug")

        cmd_line.extend(["-tlsextdebug", "-state"])

        if self.options.cert is not None:
            cmd_line.extend(["-cert", self.options.cert])
        if self.options.key is not None:
            cmd_line.extend(["-key", self.options.key])

        # Unlike s2n, OpenSSL allows us to be much more specific about which TLS
        # protocol to use.
        if self.options.protocol == Protocols.TLS13:
            cmd_line.append("-tls1_3")
        elif self.options.protocol == Protocols.TLS12:
            cmd_line.append("-tls1_2")
        elif self.options.protocol == Protocols.TLS11:
            cmd_line.append("-tls1_1")
        elif self.options.protocol == Protocols.TLS10:
            cmd_line.append("-tls1")
        elif self.options.protocol == Protocols.SSLv3:
            cmd_line.append("-ssl3")

        if self.options.cipher is not None:
            cmd_line.extend(self._cipher_to_cmdline(self.options.cipher))
            if self.options.cipher.parameters is not None:
                cmd_line.extend(["-dhparam", self.options.cipher.parameters])

        if self.options.curve is not None:
            cmd_line.extend(["-curves", str(self.options.curve)])
        if self.options.use_client_auth is True:
            # We use "Verify" instead of "verify" to require a client cert
            cmd_line.extend(["-Verify", "1"])

        if self.options.ocsp_response is not None:
            cmd_line.extend(["-status_file", self.options.ocsp_response])

        if self.options.signature_algorithm is not None:
            cmd_line.extend(["-sigalgs", self.options.signature_algorithm.name])

        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        return cmd_line

    @classmethod
    def get_name(cls, cmd_line):
        return cmd_line[1]


class SSLv3Provider(OpenSSL):
    def __init__(self, options: ProviderOptions):
        OpenSSL.__init__(self, options)
        self._override_libssl(options)

    def _override_libssl(self, options: ProviderOptions):
        install_dir = os.environ["OPENSSL_1_0_2_INSTALL_DIR"]

        override_env_vars = dict()
        override_env_vars["PATH"] = install_dir + "/bin"
        override_env_vars["LD_LIBRARY_PATH"] = install_dir + "/lib"
        options.env_overrides = override_env_vars

    @classmethod
    def supports_protocol(cls, protocol):
        if protocol is Protocols.SSLv3:
            return True
        return False


class JavaSSL(Provider):
    """
    NOTE: Only a Java SSL client has been set up. The server has not been
    implemented yet.
    """

    def __init__(self, options: ProviderOptions):
        Provider.__init__(self, options)

    @classmethod
    def get_send_marker(cls):
        return "Starting handshake"

    @classmethod
    def supports_protocol(cls, protocol):
        # https://aws.amazon.com/blogs/opensource/tls-1-0-1-1-changes-in-openjdk-and-amazon-corretto/
        if (
            protocol is Protocols.SSLv3
            or protocol is Protocols.TLS10
            or protocol is Protocols.TLS11
        ):
            return False

        return True

    @classmethod
    def supports_cipher(cls, cipher, with_curve=None):
        # Java SSL does not support CHACHA20
        if "CHACHA20" in cipher.name:
            return False

        return True

    def setup_server(self):
        pytest.skip("JavaSSL does not support server mode at this time")

    def setup_client(self):
        cmd_line = ["java", "-classpath", "bin", "SSLSocketClient"]

        if self.options.port is not None:
            cmd_line.extend([self.options.port])

        if self.options.trust_store:
            cmd_line.extend([self.options.trust_store])
        elif self.options.cert:
            cmd_line.extend([self.options.cert])
        if self.options.cipher.iana_standard_name is not None:
            cmd_line.extend([self.options.cipher.iana_standard_name])

        if self.options.protocol is not None:
            cmd_line.extend([self.options.protocol.name])
        # SSLv2ClientHello is a "protocol" for Java TLS, so we append it next to
        # the existing protocol.
        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        # Clients are always ready to connect
        self.set_provider_ready()

        return cmd_line


class BoringSSL(Provider):
    """
    NOTE: In order to focus on the general use of this framework, BoringSSL
    is not yet supported. The client works, the server has not yet been
    implemented, neither are in the default configuration.
    """

    def __init__(self, options: ProviderOptions):
        Provider.__init__(self, options)

    @classmethod
    def get_send_marker(cls):
        return "Cert issuer:"

    def setup_server(self):
        cmd_line = ["bssl", "s_server"]
        cmd_line.extend(["-accept", self.options.port])
        if self.options.cert is not None:
            cmd_line.extend(["-cert", self.options.cert])
        if self.options.key is not None:
            cmd_line.extend(["-key", self.options.key])
        if self.options.curve is not None:
            if self.options.curve == Curves.P256:
                cmd_line.extend(["-curves", "P-256"])
            elif self.options.curve == Curves.P384:
                cmd_line.extend(["-curves", "P-384"])
            elif self.options.curve == Curves.P521:
                cmd_line.extend(["-curves", "P-521"])
            elif self.options.curve == Curves.SecP256r1Kyber768Draft00:
                cmd_line.extend(["-curves", "SecP256r1Kyber768Draft00"])
            elif self.options.curve == Curves.X25519Kyber768Draft00:
                cmd_line.extend(["-curves", "X25519Kyber768Draft00"])
            elif self.options.curve == Curves.X25519:
                pytest.skip(
                    "BoringSSL does not support curve {}".format(self.options.curve)
                )

        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        return cmd_line

    def setup_client(self):
        cmd_line = ["bssl", "s_client"]
        cmd_line.extend(
            ["-connect", "{}:{}".format(self.options.host, self.options.port)]
        )
        if self.options.cert is not None:
            cmd_line.extend(["-cert", self.options.cert])
        if self.options.key is not None:
            cmd_line.extend(["-key", self.options.key])
        if self.options.cipher is not None:
            if self.options.cipher == Ciphers.TLS_CHACHA20_POLY1305_SHA256:
                cmd_line.extend(
                    ["-cipher", "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256"]
                )
        if self.options.curve is not None:
            if self.options.curve == Curves.P256:
                cmd_line.extend(["-curves", "P-256"])
            elif self.options.curve == Curves.P384:
                cmd_line.extend(["-curves", "P-384"])
            elif self.options.curve == Curves.P521:
                cmd_line.extend(["-curves", "P-521"])
            elif self.options.curve == Curves.SecP256r1Kyber768Draft00:
                cmd_line.extend(["-curves", "SecP256r1Kyber768Draft00"])
            elif self.options.curve == Curves.X25519Kyber768Draft00:
                cmd_line.extend(["-curves", "X25519Kyber768Draft00"])
            elif self.options.curve == Curves.X25519:
                pytest.skip(
                    "BoringSSL does not support curve {}".format(self.options.curve)
                )

        if self.options.extra_flags is not None:
            cmd_line.extend(self.options.extra_flags)

        # Clients are always ready to connect
        self.set_provider_ready()

        return cmd_line


class GnuTLS(Provider):
    def __init__(self, options: ProviderOptions):
        Provider.__init__(self, options)

        self.expect_stderr = True  # lgtm [py/overwritten-inherited-attribute]
        self.send_with_newline = True  # lgtm [py/overwritten-inherited-attribute]

    @staticmethod
    def cipher_to_priority_str(cipher):
        return {
            Ciphers.DHE_RSA_AES128_SHA: "DHE-RSA:+AES-128-CBC:+SHA1",
            Ciphers.DHE_RSA_AES256_SHA: "DHE-RSA:+AES-256-CBC:+SHA1",
            Ciphers.DHE_RSA_AES128_SHA256: "DHE-RSA:+AES-128-CBC:+SHA256",
            Ciphers.DHE_RSA_AES256_SHA256: "DHE-RSA:+AES-256-CBC:+SHA256",
            Ciphers.DHE_RSA_AES128_GCM_SHA256: "DHE-RSA:+AES-128-GCM:+AEAD",
            Ciphers.DHE_RSA_AES256_GCM_SHA384: "DHE-RSA:+AES-256-GCM:+AEAD",
            Ciphers.DHE_RSA_CHACHA20_POLY1305: "DHE-RSA:+CHACHA20-POLY1305:+AEAD",
            Ciphers.AES128_SHA: "RSA:+AES-128-CBC:+SHA1",
            Ciphers.AES256_SHA: "RSA:+AES-256-CBC:+SHA1",
            Ciphers.AES128_SHA256: "RSA:+AES-128-CBC:+SHA256",
            Ciphers.AES256_SHA256: "RSA:+AES-256-CBC:+SHA256",
            Ciphers.AES128_GCM_SHA256: "RSA:+AES-128-GCM:+AEAD",
            Ciphers.AES256_GCM_SHA384: "RSA:+AES-256-GCM:+AEAD",
            Ciphers.ECDHE_ECDSA_AES128_SHA: "ECDHE-ECDSA:+AES-128-CBC:+SHA1",
            Ciphers.ECDHE_ECDSA_AES256_SHA: "ECDHE-ECDSA:+AES-256-CBC:+SHA1",
            Ciphers.ECDHE_ECDSA_AES128_SHA256: "ECDHE-ECDSA:+AES-128-CBC:+SHA256",
            Ciphers.ECDHE_ECDSA_AES256_SHA384: "ECDHE-ECDSA:+AES-256-CBC:+SHA384",
            Ciphers.ECDHE_ECDSA_AES128_GCM_SHA256: "ECDHE-ECDSA:+AES-128-GCM:+AEAD",
            Ciphers.ECDHE_ECDSA_AES256_GCM_SHA384: "ECDHE-ECDSA:+AES-256-GCM:+AEAD",
            Ciphers.ECDHE_RSA_AES128_SHA: "ECDHE-RSA:+AES-128-CBC:+SHA1",
            Ciphers.ECDHE_RSA_AES256_SHA: "ECDHE-RSA:+AES-256-CBC:+SHA1",
            Ciphers.ECDHE_RSA_AES128_SHA256: "ECDHE-RSA:+AES-128-CBC:+SHA256",
            Ciphers.ECDHE_RSA_AES256_SHA384: "ECDHE-RSA:+AES-256-CBC:+SHA384",
            Ciphers.ECDHE_RSA_AES128_GCM_SHA256: "ECDHE-RSA:+AES-128-GCM:+AEAD",
            Ciphers.ECDHE_RSA_AES256_GCM_SHA384: "ECDHE-RSA:+AES-256-GCM:+AEAD",
            Ciphers.ECDHE_RSA_CHACHA20_POLY1305: "ECDHE-RSA:+CHACHA20-POLY1305:+AEAD",
        }.get(cipher)

    @staticmethod
    def protocol_to_priority_str(protocol):
        if not protocol:
            return None
        return {
            Protocols.TLS10.value: "VERS-TLS1.0",
            Protocols.TLS11.value: "VERS-TLS1.1",
            Protocols.TLS12.value: "VERS-TLS1.2",
            Protocols.TLS13.value: "VERS-TLS1.3",
        }.get(protocol.value)

    @staticmethod
    def curve_to_priority_str(curve):
        return {
            Curves.P256: "CURVE-SECP256R1",
            Curves.P384: "CURVE-SECP384R1",
            Curves.P521: "CURVE-SECP521R1",
            Curves.X25519: "CURVE-X25519",
        }.get(curve)

    @staticmethod
    def sigalg_to_priority_str(sigalg):
        return {
            Signatures.RSA_SHA1: "SIGN-RSA-SHA1",
            Signatures.RSA_SHA256: "SIGN-RSA-SHA256",
            Signatures.RSA_SHA384: "SIGN-RSA-SHA384",
            Signatures.RSA_SHA512: "SIGN-RSA-SHA512",
        }.get(sigalg)

    @classmethod
    def get_send_marker(cls):
        return "Simple Client Mode:"

    def create_priority_str(self):
        priority_str = "NONE"

        protocol_to_priority_str = self.protocol_to_priority_str(self.options.protocol)
        if protocol_to_priority_str:
            priority_str += ":+" + protocol_to_priority_str
        else:
            priority_str += ":+VERS-ALL"

        cipher_to_priority_str = self.cipher_to_priority_str(self.options.cipher)
        if cipher_to_priority_str:
            priority_str += ":+" + cipher_to_priority_str
        else:
            priority_str += ":+KX-ALL:+CIPHER-ALL:+MAC-ALL"

        curve_to_priority_str = self.curve_to_priority_str(self.options.curve)
        if curve_to_priority_str:
            priority_str += ":+" + curve_to_priority_str
        else:
            priority_str += ":+GROUP-ALL"

        sigalg_to_priority_str = self.sigalg_to_priority_str(
            self.options.signature_algorithm
        )
        if sigalg_to_priority_str:
            priority_str += ":+" + sigalg_to_priority_str
        else:
            priority_str += ":+SIGN-ALL"

        priority_str += ":+COMP-NULL"

        # A digital signature option is not included for the test RSA certs, so GnuTLS must be
        # told to use these certs regardless. The %COMPAT priority string option enables this for
        # client certificates, and the undocumented %DEBUG_ALLOW_KEY_USAGE_VIOLATIONS priority
        # string option enables this for server certificates.
        priority_str += ":%COMPAT"
        priority_str += ":%DEBUG_ALLOW_KEY_USAGE_VIOLATIONS"

        return priority_str

    def setup_client(self):
        self.set_provider_ready()

        cmd_line = [
            "gnutls-cli",
            "--port",
            str(self.options.port),
            self.options.host,
        ]

        # Most GnuTLS tests expect verbose output, so default to True.
        if self.options.verbose is not False:
            cmd_line.append("--verbose")

        if self.options.cert and self.options.key:
            cmd_line.extend(["--x509certfile", self.options.cert])
            cmd_line.extend(["--x509keyfile", self.options.key])

        priority_str = self.create_priority_str()
        cmd_line.extend(["--priority", priority_str])

        if self.options.insecure:
            cmd_line.extend(["--insecure"])

        if self.options.enable_client_ocsp:
            cmd_line.append("--ocsp")

        if self.options.record_size:
            cmd_line.extend(["--recordsize", str(self.options.record_size)])

        if self.options.extra_flags:
            cmd_line.extend(self.options.extra_flags)

        return cmd_line

    def setup_server(self):
        self.ready_to_test_marker = "Echo Server listening on"

        cmd_line = [
            "gnutls-serv",
            f"--port={self.options.port}",
            "--echo",
        ]

        if self.options.cert is not None:
            cmd_line.extend(["--x509certfile", self.options.cert])
        if self.options.key is not None:
            cmd_line.extend(["--x509keyfile", self.options.key])

        priority_str = self.create_priority_str()
        cmd_line.extend(["--priority", priority_str])

        if self.options.cipher:
            if self.options.cipher.parameters:
                cmd_line.extend(["--dhparams", self.options.cipher.parameters])

        if self.options.ocsp_response:
            cmd_line.extend(["--ocsp-response", self.options.ocsp_response])

        if self.options.use_client_auth:
            cmd_line.append("--require-client-cert")

        if self.options.extra_flags:
            cmd_line.extend(self.options.extra_flags)

        return cmd_line

    @classmethod
    def supports_protocol(cls, protocol):
        return GnuTLS.protocol_to_priority_str(protocol) is not None

    @classmethod
    def supports_cipher(cls, cipher, with_curve=None):
        return GnuTLS.cipher_to_priority_str(cipher) is not None

    @classmethod
    def supports_signature(cls, signature):
        return GnuTLS.sigalg_to_priority_str(signature) is not None


def find_files(file_glob, root_dir=".", modes=[]):
    """
    find util in python form.
    file_glob: a snippet of the filename, e.g. ".py"
    root_dir: starting point for search
    mode is an octal representation of owner/group/other, e.g.: '0o644'
    """
    result = []
    for root, dirs, files in os.walk(root_dir):
        for file in files:
            if file_glob in file:
                full_name = os.path.abspath(os.path.join(root, file))
                if len(modes) != 0:
                    try:
                        stat = oct(S_IMODE(os.stat(full_name).st_mode))
                        if stat in modes:
                            result.append(full_name)
                    except FileNotFoundError:
                        # symlinks
                        pass
                else:
                    result.append(full_name)

    return result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest

from configuration import (
    available_ports,
    TLS13_CIPHERS,
    ALL_TEST_CURVES,
    MINIMAL_TEST_CERTS,
)
from common import ProviderOptions, Protocols, data_bytes
from fixtures import managed_process  # noqa: F401
from providers import S2N, OpenSSL
from utils import invalid_test_parameters, get_parameter_name, to_bytes
from enum import Enum, auto

# Known value test vectors from https://tools.ietf.org/html/rfc8448#section-4
known_psk_identity = (
    "2c035d829359ee5ff7af4ec900000000262a6494dc486d2c8a34cb33fa90bf1b00"
    "70ad3c498883c9367c09a2be785abc55cd226097a3a982117283f82a03a143efd3"
    "ff5dd36d64e861be7fd61d2827db279cce145077d454a3664d4e6da4d29ee03725"
    "a6a4dafcd0fc67d2aea70529513e3da2677fa5906c5b3f7d8f92f228bda40dda72"
    "1470f9fbf297b5aea617646fac5c03272e970727c621a79141ef5f7de6505e5bfb"
    "c388e93343694093934ae4d357"
)
known_psk_secret = "4ecd0eb6ec3b4d87f5d6028f922ca4c5851a277fd41311c9e62d2c9492e1c4f3"

# Arbitrary test vectors
PSK_IDENTITY_LIST = [known_psk_identity, "psk_identity", "test_psk_identity"]
PSK_SECRET_LIST = [known_psk_secret, "a6dadae4567876", "a64dafcd0fc67d2a"]
PSK_IDENTITY_NO_MATCH = "PSK_IDENTITY_NO_MATCH"
PSK_SECRET_NO_MATCH = "e9492e1c"
PSK_IDENTITY_NO_MATCH_2 = "PSK_IDENTITY_NO_MATCH_2"
PSK_SECRET_NO_MATCH_2 = "c1e29493fd"

MINIMAL_TEST_CERTS_WITH_EMPTY_CERT = MINIMAL_TEST_CERTS + [None]
PSK_PROVIDERS = [OpenSSL, S2N]


class Outcome(Enum):
    psk_connection = auto()
    full_handshake = auto()
    handshake_failed = auto()


def setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg):
    return ["--psk", psk_identity + "," + psk_secret + "," + psk_hash_alg]


def setup_openssl_psk_params(psk_identity, psk_secret):
    return ["-psk_identity", psk_identity, "--psk", psk_secret]


def setup_provider_options(
    mode, port, cipher, curve, certificate, data_to_send, client_psk_params
):
    options = ProviderOptions(
        host="localhost",
        port=port,
        cipher=cipher,
        curve=curve,
        insecure=True,
        protocol=Protocols.TLS13,
        data_to_send=data_to_send,
        mode=mode,
        extra_flags=client_psk_params,
    )
    if certificate:
        options.key = certificate.key
        options.cert = certificate.cert
        options.trust_store = certificate.cert
    return options


def get_psk_hash_alg_from_cipher(cipher):
    # S2N supports only SHA256 and SHA384 PSK Hash Algorithms
    if "SHA256" in cipher.name:
        return "SHA256"
    elif "SHA384" in cipher.name:
        return "SHA384"
    else:
        return None


def skip_invalid_psk_tests(provider, psk_hash_alg):
    # If the PSK hash algorithm is None, it is not supported and we can safely skip the test case.
    if psk_hash_alg is None:
        pytest.skip()

    # In OpenSSL, PSK works only with TLS1.3 ciphersuites based on SHA256 hash algorithm which includes
    # all TLS1.3 ciphersuites supported by S2N except TLS_AES_256_GCM_SHA384.
    if provider == OpenSSL and psk_hash_alg == "SHA384":
        pytest.skip()


def validate_negotiated_psk_s2n(outcome, psk_identity, results):
    if outcome == Outcome.psk_connection:
        assert (
            to_bytes("Negotiated PSK identity: {}".format(psk_identity))
            in results.stdout
        )
    elif outcome == Outcome.full_handshake:
        assert (
            to_bytes("Negotiated PSK identity: {}".format(psk_identity))
            not in results.stdout
        )
    else:
        assert results.exit_code != 0
        assert to_bytes("Failed to negotiate: 'TLS alert received'") in results.stderr


def validate_negotiated_psk_openssl(outcome, results):
    if outcome == Outcome.psk_connection:
        assert to_bytes('extension "psk"') in results.stdout
    elif outcome == Outcome.full_handshake:
        assert (
            to_bytes("SSL_connect:SSLv3/TLS read server certificate") in results.stderr
        )
    else:
        assert to_bytes("SSL_accept:error in error") in results.stderr


def test_nothing():
    """
    Sometimes the external psk test parameters in combination with the s2n libcrypto
    results in no test cases existing. In this case, pass a nothing test to avoid
    marking the entire codebuild run as failed.
    """
    assert True


"""
Basic S2N server happy case.

Tests a single psk connection with no fallback option.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", PSK_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("psk_identity", PSK_IDENTITY_LIST, ids=get_parameter_name)
@pytest.mark.parametrize("psk_secret", PSK_SECRET_LIST, ids=get_parameter_name)
def test_s2n_server_psk_connection(
    managed_process,  # noqa: F811
    cipher,
    curve,
    protocol,
    provider,
    other_provider,
    psk_identity,
    psk_secret,
):
    port = next(available_ports)
    random_bytes = data_bytes(10)
    psk_hash_alg = get_psk_hash_alg_from_cipher(cipher)
    skip_invalid_psk_tests(provider, psk_hash_alg)

    if provider == S2N:
        client_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    else:
        client_psk_params = setup_openssl_psk_params(psk_identity, psk_secret)
    client_options = setup_provider_options(
        provider.ClientMode, port, cipher, curve, None, random_bytes, client_psk_params
    )

    server_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    server_options = setup_provider_options(
        S2N.ServerMode, port, cipher, curve, None, None, server_psk_params
    )

    server = managed_process(
        S2N, server_options, timeout=5, close_marker=str(random_bytes)
    )
    client = managed_process(provider, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()
        if provider == S2N:
            validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)
        else:
            validate_negotiated_psk_openssl(Outcome.psk_connection, results)

    for results in server.get_results():
        results.assert_success()
        validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)
        assert random_bytes in results.stdout


"""
Tests S2N server's behavior with multiple PSKs and no fallback options. 

Note that OpenSSL does not support multiple PSKs. 
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", PSK_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("psk_identity", PSK_IDENTITY_LIST, ids=get_parameter_name)
@pytest.mark.parametrize("psk_secret", PSK_SECRET_LIST, ids=get_parameter_name)
def test_s2n_server_multiple_psks(
    managed_process,  # noqa: F811
    cipher,
    curve,
    protocol,
    provider,
    other_provider,
    psk_identity,
    psk_secret,
):
    port = next(available_ports)
    random_bytes = data_bytes(10)
    psk_hash_alg = get_psk_hash_alg_from_cipher(cipher)
    skip_invalid_psk_tests(provider, psk_hash_alg)

    client_psk_params = []
    if provider == OpenSSL:
        """
        OpenSSL Provider does not support multiple PSKs in the same connection, 
        the last psk parameter is the psk parameter used in the connection. 
        """
        client_psk_params.extend(
            setup_openssl_psk_params(PSK_IDENTITY_NO_MATCH, PSK_SECRET_NO_MATCH)
        )
        client_psk_params.extend(setup_openssl_psk_params(psk_identity, psk_secret))
    else:
        client_psk_params.extend(
            setup_s2n_psk_params(
                PSK_IDENTITY_NO_MATCH, PSK_SECRET_NO_MATCH, psk_hash_alg
            )
        )
        client_psk_params.extend(
            setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
        )
    client_options = setup_provider_options(
        provider.ClientMode, port, cipher, curve, None, random_bytes, client_psk_params
    )

    server_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    server_psk_params.extend(
        setup_s2n_psk_params(
            PSK_IDENTITY_NO_MATCH_2, PSK_SECRET_NO_MATCH_2, psk_hash_alg
        )
    )
    server_options = setup_provider_options(
        S2N.ServerMode, port, cipher, curve, None, None, server_psk_params
    )

    server = managed_process(
        S2N, server_options, timeout=5, close_marker=str(random_bytes)
    )
    client = managed_process(provider, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()
        if provider == S2N:
            validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)
        else:
            validate_negotiated_psk_openssl(Outcome.psk_connection, results)

    for results in server.get_results():
        results.assert_success()
        validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)
        assert random_bytes in results.stdout


"""
Tests S2N Server's fallback to full handshake.

Verify S2N Server's fallback behavior with an invalid PSK parameter and a valid certificate as the input.

Note that S2N Server succeeds with a full handshake when an invalid PSK parameter and an empty 
certificate is provided as the input, as S2N Server uses a default certificate if a certificate is not provided 
as the input.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", PSK_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("psk_identity", PSK_IDENTITY_LIST, ids=get_parameter_name)
@pytest.mark.parametrize("psk_secret", PSK_SECRET_LIST, ids=get_parameter_name)
@pytest.mark.parametrize(
    "certificate", MINIMAL_TEST_CERTS_WITH_EMPTY_CERT, ids=get_parameter_name
)
def test_s2n_server_full_handshake(
    managed_process,  # noqa: F811
    cipher,
    curve,
    protocol,
    provider,
    other_provider,
    psk_identity,
    psk_secret,
    certificate,
):
    port = next(available_ports)
    random_bytes = data_bytes(10)
    psk_hash_alg = get_psk_hash_alg_from_cipher(cipher)
    skip_invalid_psk_tests(provider, psk_hash_alg)

    if provider == S2N:
        client_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    else:
        client_psk_params = setup_openssl_psk_params(psk_identity, psk_secret)
    client_options = setup_provider_options(
        provider.ClientMode,
        port,
        cipher,
        curve,
        certificate,
        random_bytes,
        client_psk_params,
    )

    server_psk_params = setup_s2n_psk_params(
        PSK_IDENTITY_NO_MATCH, PSK_SECRET_NO_MATCH, psk_hash_alg
    )
    server_options = setup_provider_options(
        S2N.ServerMode, port, cipher, curve, certificate, None, server_psk_params
    )

    server = managed_process(
        S2N, server_options, timeout=5, close_marker=str(random_bytes)
    )
    client = managed_process(provider, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()
        if provider == S2N:
            validate_negotiated_psk_s2n(Outcome.full_handshake, psk_identity, results)
        else:
            validate_negotiated_psk_openssl(Outcome.full_handshake, results)

    for results in server.get_results():
        results.assert_success()
        validate_negotiated_psk_s2n(
            Outcome.full_handshake, PSK_IDENTITY_NO_MATCH, results
        )
        assert random_bytes in results.stdout


"""
Basic S2N client happy case.

Tests a single psk connection with no fallback option.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", PSK_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("psk_identity", PSK_IDENTITY_LIST, ids=get_parameter_name)
@pytest.mark.parametrize("psk_secret", PSK_SECRET_LIST, ids=get_parameter_name)
def test_s2n_client_psk_connection(
    managed_process,  # noqa: F811
    cipher,
    curve,
    protocol,
    provider,
    other_provider,
    psk_identity,
    psk_secret,
):
    port = next(available_ports)
    random_bytes = data_bytes(10)
    psk_hash_alg = get_psk_hash_alg_from_cipher(cipher)
    skip_invalid_psk_tests(provider, psk_hash_alg)

    client_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    client_options = setup_provider_options(
        S2N.ClientMode, port, cipher, curve, None, random_bytes, client_psk_params
    )

    if provider == S2N:
        server_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    else:
        server_psk_params = setup_openssl_psk_params(psk_identity, psk_secret)
        server_psk_params += ["-nocert"]
    server_options = setup_provider_options(
        provider.ServerMode, port, cipher, curve, None, None, server_psk_params
    )

    server = managed_process(
        provider, server_options, timeout=5, close_marker=str(random_bytes)
    )
    client = managed_process(S2N, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()
        validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)

    for results in server.get_results():
        results.assert_success()
        if provider == S2N:
            validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)
        else:
            validate_negotiated_psk_openssl(Outcome.psk_connection, results)
        assert random_bytes in results.stdout


"""
Tests S2N client's behavior with multiple PSKs and no fallback option. 

Note that OpenSSL does not support multiple PSKs. 
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", PSK_PROVIDERS, ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("psk_identity", PSK_IDENTITY_LIST, ids=get_parameter_name)
@pytest.mark.parametrize("psk_secret", PSK_SECRET_LIST, ids=get_parameter_name)
def test_s2n_client_multiple_psks(
    managed_process,  # noqa: F811
    cipher,
    curve,
    protocol,
    provider,
    other_provider,
    psk_identity,
    psk_secret,
):
    port = next(available_ports)
    random_bytes = data_bytes(10)
    psk_hash_alg = get_psk_hash_alg_from_cipher(cipher)
    skip_invalid_psk_tests(provider, psk_hash_alg)

    client_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    client_psk_params.extend(
        setup_s2n_psk_params(PSK_IDENTITY_NO_MATCH, PSK_SECRET_NO_MATCH, psk_hash_alg)
    )
    client_options = setup_provider_options(
        S2N.ClientMode, port, cipher, curve, None, random_bytes, client_psk_params
    )

    server_psk_params = []
    if provider == OpenSSL:
        """
        OpenSSL Provider does not support multiple PSKs in the same connection, 
        the last psk params is the final psk used in the connection. 
        """
        server_psk_params.extend(
            setup_openssl_psk_params(PSK_IDENTITY_NO_MATCH_2, PSK_SECRET_NO_MATCH_2)
        )
        server_psk_params.extend(setup_openssl_psk_params(psk_identity, psk_secret))
        server_psk_params += ["-nocert"]
    else:
        server_psk_params.extend(
            setup_s2n_psk_params(
                PSK_IDENTITY_NO_MATCH_2, PSK_SECRET_NO_MATCH_2, psk_hash_alg
            )
        )
        server_psk_params.extend(
            setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
        )
    server_options = setup_provider_options(
        provider.ServerMode, port, cipher, curve, None, None, server_psk_params
    )

    server = managed_process(
        provider, server_options, timeout=5, close_marker=str(random_bytes)
    )
    client = managed_process(S2N, client_options, timeout=5)

    for results in client.get_results():
        results.assert_success()
        validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)

    for results in server.get_results():
        results.assert_success()
        if provider == S2N:
            validate_negotiated_psk_s2n(Outcome.psk_connection, psk_identity, results)
        else:
            validate_negotiated_psk_openssl(Outcome.psk_connection, results)
        assert random_bytes in results.stdout


"""
S2N Client fails to succeed with a handshake when an invalid PSK parameter
and an invalid certificate is provided as the input.

Note that we cannot use S2N Server as a provider input for this test as S2N Server 
uses a default certificate if a certificate is not provided as the input.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("cipher", TLS13_CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("curve", ALL_TEST_CURVES, ids=get_parameter_name)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize("provider", [OpenSSL], ids=get_parameter_name)
@pytest.mark.parametrize("psk_identity", PSK_IDENTITY_LIST, ids=get_parameter_name)
@pytest.mark.parametrize("psk_secret", PSK_SECRET_LIST, ids=get_parameter_name)
def test_s2n_client_psk_handshake_failure(
    managed_process,  # noqa: F811
    cipher,
    curve,
    protocol,
    provider,
    psk_identity,
    psk_secret,
):
    port = next(available_ports)
    random_bytes = data_bytes(10)
    psk_hash_alg = get_psk_hash_alg_from_cipher(cipher)
    skip_invalid_psk_tests(provider, psk_hash_alg)

    client_psk_params = setup_s2n_psk_params(psk_identity, psk_secret, psk_hash_alg)
    client_options = setup_provider_options(
        S2N.ClientMode, port, cipher, curve, None, random_bytes, client_psk_params
    )

    server_psk_params = setup_openssl_psk_params(
        PSK_IDENTITY_NO_MATCH, PSK_SECRET_NO_MATCH
    )
    server_psk_params += ["-nocert"]
    server_options = setup_provider_options(
        provider.ServerMode, port, cipher, curve, None, None, server_psk_params
    )

    server = managed_process(
        provider, server_options, timeout=5, close_marker=str(random_bytes)
    )
    client = managed_process(S2N, client_options, timeout=5)

    for results in client.get_results():
        assert to_bytes("Failed to negotiate: 'TLS alert received'") in results.stderr
        validate_negotiated_psk_s2n(Outcome.handshake_failed, psk_identity, results)

    for results in server.get_results():
        assert to_bytes("SSL_accept:error in error") in results.stderr
        validate_negotiated_psk_openssl(Outcome.handshake_failed, results)
        assert random_bytes not in results.stdout



#!/usr/bin/env python3
import os
import shutil
import tempfile
import subprocess
import re


def run_command(cmd, cwd=None, check=True):
    """Run a command and return its output."""
    print(f"Running: {cmd}")
    result = subprocess.run(cmd, shell=True, check=check, text=True,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
    return result.stdout.strip(), result.returncode


def main():
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp()
    print(f"Created temporary directory: {temp_dir}")

    try:
        # 1. Clone the cJSON repository
        _, _ = run_command(
            f"git clone https://github.com/DaveGamble/cJSON.git {temp_dir}")

        # 2. Get the latest release tag
        tags_output, _ = run_command("git tag -l", cwd=temp_dir)
        tags = tags_output.split('\n')
        # Filter for version tags (e.g., v1.7.15)
        version_tags = [tag for tag in tags if re.match(
            r'^v\d+\.\d+\.\d+$', tag)]
        # Sort by version number
        version_tags.sort(key=lambda v: [int(x) for x in v[1:].split('.')])
        latest_tag = version_tags[-1]
        latest_version = latest_tag[1:]  # Remove 'v' prefix

        print(f"Latest cJSON version: {latest_version}")

        # Checkout the latest tag
        _, _ = run_command(f"git checkout {latest_tag}", cwd=temp_dir)

        # 3. Copy cJSON.h and cJSON.c to ./source/external/
        base_dir = os.path.join(os.path.dirname(
            __file__), "..")
        source_dir = os.path.join(base_dir, "source", "external")
        shutil.copy(os.path.join(temp_dir, "cJSON.h"),
                    os.path.join(source_dir, "cJSON.h"))
        shutil.copy(os.path.join(temp_dir, "cJSON.c"),
                    os.path.join(source_dir, "cJSON.c"))

        print(f"Copied cJSON.h and cJSON.c to {source_dir}")

        # 4. Update the version number in THIRD-PARTY-LICENSES.txt
        license_path = os.path.join(base_dir, "THIRD-PARTY-LICENSES.txt")
        with open(license_path, 'r') as f:
            content = f.read()

        # Update the version number using regex for more accurate replacement
        pattern = r'\*\* cJSON; version \d+\.\d+\.\d+ --'
        replacement = f'** cJSON; version {latest_version} --'
        content = re.sub(pattern, replacement, content)

        with open(license_path, 'w') as f:
            f.write(content)

        print(f"Updated cJSON to version {latest_version}")

        # 5. Apply the changes from commit f005fdc06c34fdd663031661eff5c5575843e998
        # https://github.com/awslabs/aws-c-common/pull/1211/commits/f005fdc06c34fdd663031661eff5c5575843e998
        # it applies the amazon edit to cJSON.h and cJSON.c
        _, returncode = run_command(
            f"git add *", cwd=source_dir, check=False)
        if returncode == 0:
            _, returncode = run_command(
                f"git commit -m \"update cjson to {latest_version}\"", cwd=base_dir, check=False)
            if returncode == 0:
                _, returncode = run_command(
                    f"git cherry-pick f005fdc06c34fdd663031661eff5c5575843e998", cwd=base_dir, check=False)
                if returncode == 0:
                    print(
                        "Successfully applied Amazon modifications from commit f005fdc06c34fdd663031661eff5c5575843e998")
                else:
                    print(
                        "\nWARNING: Failed to apply changes from commit f005fdc06c34fdd663031661eff5c5575843e998")
                    print(
                        "You will need to manually apply these changes. Please run the following command:")
                    print(f"  git cherry-pick f005fdc06c34fdd663031661eff5c5575843e998")
                    print("If there are conflicts, resolve them and then run:")
                    print("  git cherry-pick --continue")
            else:
                print("\nWARNING: Failed to commit the updated cJSON files")
                print("You will need to manually commit and then apply the changes from commit f005fdc06c34fdd663031661eff5c5575843e998")
        else:
            print("\nWARNING: Failed to add the updated cJSON files to git")
            print("You will need to manually add, commit, and then apply the changes from commit f005fdc06c34fdd663031661eff5c5575843e998")

    finally:
        # Clean up the temporary directory
        shutil.rmtree(temp_dir)
        print(f"Removed temporary directory: {temp_dir}")


if __name__ == "__main__":
    main()
